  # -*- coding: utf-8 -*-
"""Tests xonsh tools."""
import datetime as dt
import os
import pathlib
import stat
from tempfile import TemporaryDirectory

import pytest

from xonsh.platform import ON_WINDOWS
from xonsh.lexer import Lexer

from xonsh.tools import (
    EnvPath, always_false, always_true, argvquote,
    bool_or_int_to_str, bool_to_str, check_for_partial_string,
    dynamic_cwd_tuple_to_str, ensure_slice, ensure_string,
    env_path_to_str, escape_windows_cmd_string, executables_in,
    expand_case_matching, expand_path, find_next_break, is_bool, is_bool_or_int,
    is_callable, is_dynamic_cwd_width, is_env_path, is_float, is_int, is_logfile_opt,
    is_string_or_callable, logfile_opt_to_str, str_to_env_path, is_string,
    subexpr_from_unbalanced, subproc_toks, to_bool, to_bool_or_int,
    to_dynamic_cwd_tuple, to_logfile_opt, pathsep_to_set, set_to_pathsep,
    is_string_seq, pathsep_to_seq, seq_to_pathsep, is_nonstring_seq_of_strings,
    pathsep_to_upper_seq, seq_to_upper_pathsep, expandvars, is_int_as_str, is_slice_as_str,
    ensure_timestamp, get_portions, is_balanced, subexpr_before_unbalanced,
    swap_values, get_logical_line, replace_logical_line, check_quotes,
    )
from xonsh.environ import Env

from tools import skip_if_on_windows, skip_if_on_unix

LEXER = Lexer()
LEXER.build()

INDENT = '    '

TOOLS_ENV = {'EXPAND_ENV_VARS': True, 'XONSH_ENCODING_ERRORS':'strict'}
ENCODE_ENV_ONLY = {'XONSH_ENCODING_ERRORS': 'strict'}
PATHEXT_ENV = {'PATHEXT': ['.COM', '.EXE', '.BAT']}

def test_subproc_toks_x():
    exp = '![x]'
    obs = subproc_toks('x', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_l():
    exp = '![ls -l]'
    obs = subproc_toks('ls -l', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_git():
    s = 'git commit -am "hello doc"'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_git_semi():
    s = 'git commit -am "hello doc"'
    exp = '![{0}];'.format(s)
    obs = subproc_toks(s + ';', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_git_nl():
    s = 'git commit -am "hello doc"'
    exp = '![{0}]\n'.format(s)
    obs = subproc_toks(s + '\n', lexer=LEXER, returnline=True)
    assert exp == obs


def test_bash_macro():
    s = 'bash -c ! export var=42; echo $var'
    exp = '![{0}]\n'.format(s)
    obs = subproc_toks(s + '\n', lexer=LEXER, returnline=True)
    assert exp == obs


def test_python_macro():
    s = 'python -c ! import os; print(os.path.abspath("/"))'
    exp = '![{0}]\n'.format(s)
    obs = subproc_toks(s + '\n', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls():
    s = 'ls -l'
    exp = INDENT + '![{0}]'.format(s)
    obs = subproc_toks(INDENT + s, mincol=len(INDENT), lexer=LEXER,
                       returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_nl():
    s = 'ls -l'
    exp = INDENT + '![{0}]\n'.format(s)
    obs = subproc_toks(INDENT + s + '\n', mincol=len(INDENT), lexer=LEXER,
                       returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_no_min():
    s = 'ls -l'
    exp = INDENT + '![{0}]'.format(s)
    obs = subproc_toks(INDENT + s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_no_min_nl():
    s = 'ls -l'
    exp = INDENT + '![{0}]\n'.format(s)
    obs = subproc_toks(INDENT + s + '\n', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_no_min_semi():
    s = 'ls'
    exp = INDENT + '![{0}];'.format(s)
    obs = subproc_toks(INDENT + s + ';', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_no_min_semi_nl():
    s = 'ls'
    exp = INDENT + '![{0}];\n'.format(s)
    obs = subproc_toks(INDENT + s + ';\n', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_comment():
    s = 'ls -l'
    com = '  # lets list'
    exp = '![{0}]{1}'.format(s, com)
    obs = subproc_toks(s + com, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_42_comment():
    s = 'ls 42'
    com = '  # lets list'
    exp = '![{0}]{1}'.format(s, com)
    obs = subproc_toks(s + com, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_str_comment():
    s = 'ls "wakka"'
    com = '  # lets list'
    exp = '![{0}]{1}'.format(s, com)
    obs = subproc_toks(s + com, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_comment():
    ind = '    '
    s = 'ls -l'
    com = '  # lets list'
    exp = '{0}![{1}]{2}'.format(ind, s, com)
    obs = subproc_toks(ind + s + com, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_indent_ls_str():
    ind = '    '
    s = 'ls "wakka"'
    com = '  # lets list'
    exp = '{0}![{1}]{2}'.format(ind, s, com)
    obs = subproc_toks(ind + s + com, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_l_semi_ls_first():
    lsdl = 'ls -l'
    ls = 'ls'
    s = '{0}; {1}'.format(lsdl, ls)
    exp = '![{0}]; {1}'.format(lsdl, ls)
    obs = subproc_toks(s, lexer=LEXER, maxcol=6, returnline=True)
    assert exp == obs


def test_subproc_toks_ls_l_semi_ls_second():
    lsdl = 'ls -l'
    ls = 'ls'
    s = '{0}; {1}'.format(lsdl, ls)
    exp = '{0}; ![{1}]'.format(lsdl, ls)
    obs = subproc_toks(s, lexer=LEXER, mincol=7, returnline=True)
    assert exp == obs


def test_subproc_toks_hello_mom_first():
    fst = "echo 'hello'"
    sec = "echo 'mom'"
    s = '{0}; {1}'.format(fst, sec)
    exp = '![{0}]; {1}'.format(fst, sec)
    obs = subproc_toks(s, lexer=LEXER, maxcol=len(fst)+1, returnline=True)
    assert exp == obs


def test_subproc_toks_hello_mom_second():
    fst = "echo 'hello'"
    sec = "echo 'mom'"
    s = '{0}; {1}'.format(fst, sec)
    exp = '{0}; ![{1}]'.format(fst, sec)
    obs = subproc_toks(s, lexer=LEXER, mincol=len(fst), returnline=True)
    assert exp == obs


def test_subproc_toks_hello_bad_leading_single_quotes():
    obs = subproc_toks('echo "hello', lexer=LEXER, returnline=True)
    assert obs is None


def test_subproc_toks_hello_bad_trailing_single_quotes():
    obs = subproc_toks('echo hello"', lexer=LEXER, returnline=True)
    assert obs is None


def test_subproc_toks_hello_bad_leading_triple_quotes():
    obs = subproc_toks('echo """hello', lexer=LEXER, returnline=True)
    assert obs is None


def test_subproc_toks_hello_bad_trailing_triple_quotes():
    obs = subproc_toks('echo hello"""', lexer=LEXER, returnline=True)
    assert obs is None


def test_subproc_toks_hello_mom_triple_quotes_nl():
    s = 'echo """hello\nmom"""'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_comment():
    exp = None
    obs = subproc_toks('# I am a comment', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_not():
    exp = 'not ![echo mom]'
    obs = subproc_toks('not echo mom', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_paren():
    exp = '(![echo mom])'
    obs = subproc_toks('(echo mom)', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_paren_ws():
    exp = '(![echo mom])  '
    obs = subproc_toks('(echo mom)  ', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_not_paren():
    exp = 'not (![echo mom])'
    obs = subproc_toks('not (echo mom)', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_and_paren():
    exp = 'True and (![echo mom])'
    obs = subproc_toks('True and (echo mom)', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_paren_and_paren():
    exp = '(![echo a]) and (echo b)'
    obs = subproc_toks('(echo a) and (echo b)', maxcol=9, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_semicolon_only():
    exp = None
    obs = subproc_toks(';', lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_pyeval():
    s = 'echo @(1+1)'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_twopyeval():
    s = 'echo @(1+1) @(40 + 2)'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_pyeval_parens():
    s = 'echo @(1+1)'
    inp = '({0})'.format(s)
    exp = '(![{0}])'.format(s)
    obs = subproc_toks(inp, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_twopyeval_parens():
    s = 'echo @(1+1) @(40+2)'
    inp = '({0})'.format(s)
    exp = '(![{0}])'.format(s)
    obs = subproc_toks(inp, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_pyeval_nested():
    s = 'echo @(min(1, 42))'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_pyeval_nested_parens():
    s = 'echo @(min(1, 42))'
    inp = '({0})'.format(s)
    exp = '(![{0}])'.format(s)
    obs = subproc_toks(inp, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_capstdout():
    s = 'echo $(echo bat)'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_capproc():
    s = 'echo !(echo bat)'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_pyeval_redirect():
    s = 'echo @("foo") > bar'
    inp = '{0}'.format(s)
    exp = '![{0}]'.format(s)
    obs = subproc_toks(inp, lexer=LEXER, returnline=True)
    assert exp == obs


def test_subproc_toks_greedy_parens():
    s = '(sort)'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True, greedy=True)
    assert exp == obs


def test_subproc_toks_greedy_parens_inp():
    s = '(sort) < input.txt'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True, greedy=True)
    assert exp == obs


def test_subproc_toks_greedy_parens_statements():
    s = '(echo "abc"; sleep 1; echo "def")'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True, greedy=True)
    assert exp == obs


def test_subproc_toks_greedy_parens_statements_with_grep():
    s = '(echo "abc"; sleep 1; echo "def") | grep'
    exp = '![{0}]'.format(s)
    obs = subproc_toks(s, lexer=LEXER, returnline=True, greedy=True)
    assert exp == obs


LOGICAL_LINE_CASES = [
("""x = 14 + 2""", 0, 'x = 14 + 2', 1),
("""x = \\
14 \\
+ 2
""", 0, 'x = 14 + 2', 3),
("""y = 16
14 \\
+ 2
""", 1, '14 + 2', 2),
('''x = """wow
mom"""
''', 0, 'x = """wow\nmom"""', 2),
# test from start
("echo --option1 value1 \\\n"
 "     --option2 value2 \\\n"
 "     --optionZ valueZ",
0, "echo --option1 value1      --option2 value2      --optionZ valueZ", 3),
# test from second line
("echo --option1 value1 \\\n"
 "     --option2 value2 \\\n"
 "     --optionZ valueZ",
1, "echo --option1 value1      --option2 value2      --optionZ valueZ", 3),
]

@pytest.mark.parametrize('src, idx, exp_line, exp_n', LOGICAL_LINE_CASES)
def test_get_logical_line(src, idx, exp_line, exp_n):
    lines = src.splitlines()
    line, n, start = get_logical_line(lines, idx)
    assert exp_line == line
    assert exp_n == n


@pytest.mark.parametrize('src, idx, exp_line, exp_n', LOGICAL_LINE_CASES)
def test_replace_logical_line(src, idx, exp_line, exp_n):
    lines = src.splitlines()
    logical = exp_line
    while idx > 0 and lines[idx-1].endswith('\\'):
        idx -= 1
    replace_logical_line(lines, logical, idx, exp_n)
    exp = src.replace('\\\n', '').strip()
    obs = '\n'.join(lines).replace('\\\n', '').strip()
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('f(1,10),x.y', True),
    ('"x"', True),
    ("'y'", True),
    ('b"x"', True),
    ("r'y'", True),
    ('"""hello\nmom"""', True),
])
def test_check_quotes(inp, exp):
    obs = check_quotes(inp)
    assert exp is obs


@pytest.mark.parametrize('inp', [
    'f(1,10),x.y',
])
def test_is_balanced_parens(inp):
    obs = is_balanced(inp, '(', ')')
    assert obs


@pytest.mark.parametrize('inp', [
    'f(x.',
    'f(1,x.'
    'f((1,10),x.y',
])
def test_is_not_balanced_parens(inp):
    obs = is_balanced(inp, '(', ')')
    assert not obs


@pytest.mark.parametrize('inp, exp', [
    ('f(x.', 'x.'),
    ('f(1,x.', 'x.'),
    ('f((1,10),x.y', 'x.y'),
])
def test_subexpr_from_unbalanced_parens(inp, exp):
    obs = subexpr_from_unbalanced(inp, '(', ')')
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('f(x.', 'f'),
    ('f(1,x.', 'f'),
    ('f((1,10),x.y', 'f'),
    ('wakka().f((1,10),x.y', '.f'),
    ('wakka(f((1,10),x.y', 'f'),
    ('wakka(jawakka().f((1,10),x.y', '.f'),
    ('wakka(jawakka().f((1,10),x.y)', 'wakka'),
])
def test_subexpr_before_unbalanced_parens(inp, exp):
    obs = subexpr_before_unbalanced(inp, '(', ')')
    assert exp == obs


@pytest.mark.parametrize('line, mincol, exp', [
    ('ls && echo a', 0, 4),
    ('ls && echo a', 6, None),
    ('ls && echo a || echo b', 6, 14),
    ('(ls) && echo a', 1, 4),
    ('not ls && echo a', 0, 8),
    ('not (ls) && echo a', 0, 8),
    ('bash -c ! export var=42; echo $var', 0, 35),
    ('python -c ! import os; print(os.path.abspath("/"))', 0, 51),
])
def test_find_next_break(line, mincol, exp):
    obs = find_next_break(line, mincol=mincol, lexer=LEXER)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (42, True),
    (42.0, False),
    ('42', False),
    ('42.0', False),
    ([42], False),
    ([], False),
    (None, False),
    ('', False)
])
def test_is_int(inp, exp):
    obs = is_int(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (42.0, True),
    (42.000101010010101010101001010101010001011100001101101011100, True),
    (42, False),
    ('42', False),
    ('42.0', False),
    ([42], False),
    ([], False),
    (None, False),
    ('', False),
    (False, False),
    (True, False),
])
def test_is_float(inp, exp):
    obs = is_float(inp)
    assert exp == obs


def test_is_string_true():
    assert is_string('42.0')

def test_is_string_false():
    assert not is_string(42.0)


def test_is_callable_true():
    assert is_callable(lambda: 42.0)


def test_is_callable_false():
    assert not is_callable(42.0)


@pytest.mark.parametrize('inp', ['42.0', lambda: 42.0])
def test_is_string_or_callable_true(inp):
    assert is_string_or_callable(inp)


def test_is_string_or_callable_false():
    assert not is_string(42.0)


@pytest.mark.parametrize('inp', [42, '42'])
def test_always_true(inp):
    assert always_true(inp)


@pytest.mark.parametrize('inp', [42, '42'])
def test_always_false(inp):
    assert not always_false(inp)


@pytest.mark.parametrize('inp, exp', [(42, '42'), ('42', '42'),])
def test_ensure_string(inp, exp):
    obs = ensure_string(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('', set()),
    ('a', {'a'}),
    (os.pathsep.join(['a', 'b']), {'a', 'b'}),
    (os.pathsep.join(['a', 'b', 'c']), {'a', 'b', 'c'}),
])
def test_pathsep_to_set(inp, exp):
    obs = pathsep_to_set(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (set(), ''),
    ({'a'}, 'a'),
    ({'a', 'b'}, os.pathsep.join(['a', 'b'])),
    ({'a', 'b', 'c'}, os.pathsep.join(['a', 'b', 'c'])),
])
def test_set_to_pathsep(inp, exp):
    obs = set_to_pathsep(inp, sort=(len(inp) > 1))
    assert exp == obs


@pytest.mark.parametrize('inp', ['42.0', ['42.0']])
def test_is_string_seq_true(inp):
    assert is_string_seq(inp)


def test_is_string_seq_false():
    assert not is_string_seq([42.0])


def test_is_nonstring_seq_of_strings_true():
    assert is_nonstring_seq_of_strings(['42.0'])


def test_is_nonstring_seq_of_strings_false():
    assert not is_nonstring_seq_of_strings([42.0])


@pytest.mark.parametrize('inp, exp', [
    ('', []),
    ('a', ['a']),
    (os.pathsep.join(['a', 'b']), ['a', 'b']),
    (os.pathsep.join(['a', 'b', 'c']), ['a', 'b', 'c']),
])
def test_pathsep_to_seq(inp, exp):
    obs = pathsep_to_seq(inp)
    assert exp == obs

@pytest.mark.parametrize('inp, exp', [
    ([], ''),
    (['a'], 'a'),
    (['a', 'b'], os.pathsep.join(['a', 'b'])),
    (['a', 'b', 'c'], os.pathsep.join(['a', 'b', 'c'])),
])
def test_seq_to_pathsep(inp, exp):
    obs = seq_to_pathsep(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('', []),
    ('a', ['A']),
    (os.pathsep.join(['a', 'B']), ['A', 'B']),
    (os.pathsep.join(['A', 'b', 'c']), ['A', 'B', 'C']),
])
def test_pathsep_to_upper_seq(inp, exp):
    obs = pathsep_to_upper_seq(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
        ([], ''),
        (['a'], 'A'),
        (['a', 'b'], os.pathsep.join(['A', 'B'])),
        (['a', 'B', 'c'], os.pathsep.join(['A', 'B', 'C'])),
        ])
def test_seq_to_upper_pathsep(inp, exp):
    obs = seq_to_upper_pathsep(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('/home/wakka', False),
    (['/home/jawaka'], False),
    (EnvPath(['/home/jawaka']), True),
    (EnvPath(['jawaka']), True),
    (EnvPath(b'jawaka:wakka'), True),
])
def test_is_env_path(inp, exp):
    obs = is_env_path(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('/home/wakka', ['/home/wakka']),
    ('/home/wakka' + os.pathsep + '/home/jawaka',
     ['/home/wakka', '/home/jawaka']),
    (b'/home/wakka', ['/home/wakka']),
])
def test_str_to_env_path(inp, exp):
    obs = str_to_env_path(inp)
    assert exp == obs.paths


@pytest.mark.parametrize('inp, exp', [
    (['/home/wakka'], '/home/wakka'),
    (['/home/wakka', '/home/jawaka'],
     '/home/wakka' + os.pathsep + '/home/jawaka'),
])
def test_env_path_to_str(inp, exp):
    obs = env_path_to_str(inp)
    assert exp == obs


# helper
def expand(path):
    return os.path.expanduser(os.path.expandvars(path))

@pytest.mark.parametrize('env', [TOOLS_ENV, ENCODE_ENV_ONLY])
@pytest.mark.parametrize('inp, exp', [
    ('xonsh_dir', 'xonsh_dir'),
    ('.', '.'),
    ('../', '../'),
    ('~/', '~/'),
    (b'~/../', '~/../'),
])
def test_env_path_getitem(inp, exp, xonsh_builtins, env):
    xonsh_builtins.__xonsh_env__ = env
    obs = EnvPath(inp)[0] # call to __getitem__
    if env.get('EXPAND_ENV_VARS'):
        assert expand(exp) == obs
    else:
        assert exp == obs


@pytest.mark.parametrize('env', [TOOLS_ENV, ENCODE_ENV_ONLY])
@pytest.mark.parametrize('inp, exp', [
    (os.pathsep.join(['xonsh_dir', '../', '.', '~/']),
     ['xonsh_dir', '../', '.', '~/']),
    ('/home/wakka' + os.pathsep + '/home/jakka' + os.pathsep + '~/',
     ['/home/wakka', '/home/jakka', '~/'])
])
def test_env_path_multipath(inp, exp, xonsh_builtins, env):
    # cases that involve path-separated strings
    xonsh_builtins.__xonsh_env__ = env
    if env == TOOLS_ENV:
        obs = [i for i in EnvPath(inp)]
        assert [expand(i) for i in exp] == obs
    else:
        obs = [i for i in EnvPath(inp)]
        assert [i for i in exp] == obs


@pytest.mark.parametrize('inp, exp', [
    (pathlib.Path('/home/wakka'), ['/home/wakka'.replace('/', os.sep)]),
    (pathlib.Path('~/'), ['~']),
    (pathlib.Path('.'), ['.']),
    (['/home/wakka', pathlib.Path('/home/jakka'), '~/'],
     ['/home/wakka', '/home/jakka'.replace('/', os.sep), '~/']),
    (['/home/wakka', pathlib.Path('../'), '../'],
     ['/home/wakka', '..', '../']),
    (['/home/wakka', pathlib.Path('~/'), '~/'],
     ['/home/wakka', '~', '~/']),
])
def test_env_path_with_pathlib_path_objects(inp, exp, xonsh_builtins):
    xonsh_builtins.__xonsh_env__ = TOOLS_ENV
    # iterate over EnvPath to acquire all expanded paths
    obs = [i for i in EnvPath(inp)]
    assert [expand(i) for i in exp] == obs

@pytest.mark.parametrize('inp', ['42.0', [42.0]])
def test_is_nonstring_seq_of_strings_false(inp):
    assert not is_nonstring_seq_of_strings(inp)


# helper
def mkpath(*paths):
    """Build os-dependent paths properly."""
    return os.sep + os.sep.join(paths)


@pytest.mark.parametrize('inp, exp', [
    ([mkpath('home', 'wakka'),
      mkpath('home', 'jakka'),
      mkpath('home', 'yakka')],
     [mkpath('home', 'wakka'),
      mkpath('home', 'jakka')])
])
def test_env_path_slice_get_all_except_last_element(inp, exp):
    obs = EnvPath(inp)[:-1]
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
        ([mkpath('home', 'wakka'),
          mkpath('home', 'jakka'),
          mkpath('home', 'yakka')],
         [mkpath('home', 'jakka'),
          mkpath('home', 'yakka')])
])
def test_env_path_slice_get_all_except_first_element(inp, exp):
    obs = EnvPath(inp)[1:]
    assert exp == obs


@pytest.mark.parametrize('inp, exp_a, exp_b', [
        ([mkpath('home', 'wakka'),
          mkpath('home', 'jakka'),
          mkpath('home', 'yakka'),
          mkpath('home', 'takka')],
         [mkpath('home', 'wakka'),
          mkpath('home', 'yakka')],
         [mkpath('home', 'jakka'),
          mkpath('home', 'takka')])
])
def test_env_path_slice_path_with_step(inp, exp_a, exp_b):
    obs_a = EnvPath(inp)[0::2]
    assert exp_a == obs_a
    obs_b = EnvPath(inp)[1::2]
    assert exp_b == obs_b


@pytest.mark.parametrize('inp, exp', [
        ([mkpath('home', 'wakka'),
          mkpath('home', 'xakka'),
          mkpath('other', 'zakka'),
          mkpath('another', 'akka'),
          mkpath('home', 'bakka')],
         [mkpath('other', 'zakka'),
          mkpath('another', 'akka')])
])
def test_env_path_keep_only_non_home_paths(inp, exp):
    obs = EnvPath(inp)[2:4]
    assert exp == obs


@pytest.mark.parametrize('inp', [True, False])
def test_is_bool_true(inp):
    assert True == is_bool(inp)


@pytest.mark.parametrize('inp', [1, 'yooo hooo!'])
def test_is_bool_false(inp):
    assert False == is_bool(inp)


@pytest.mark.parametrize('inp, exp', [
    (True, True),
    (False, False),
    (None, False),
    ('', False),
    ('0', False),
    ('False', False),
    ('NONE', False),
    ('TRUE', True),
    ('1', True),
    (0, False),
    (1, True),
])
def test_to_bool(inp, exp):
    obs = to_bool(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [(True, '1'), (False, '')])
def test_bool_to_str(inp, exp):
    assert bool_to_str(inp) == exp


@pytest.mark.parametrize('inp, exp', [
    (True, True),
    (False, True),
    (1, True),
    (0, True),
    ('Yolo', False),
    (1.0, False),
])
def test_is_bool_or_int(inp, exp):
    obs = is_bool_or_int(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (True, True),
    (False, False),
    (1, 1),
    (0, 0),
    ('', False),
    (0.0, False),
    (1.0, True),
    ('T', True),
    ('f', False),
    ('0', 0),
    ('10', 10),
])
def test_to_bool_or_int(inp, exp):
    obs = to_bool_or_int(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (True, '1'),
    (False, ''),
    (1, '1'),
    (0, '0'),
])
def test_bool_or_int_to_str(inp, exp):
    obs = bool_or_int_to_str(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
        (42, slice(42, 43)),
        (0, slice(0, 1)),
        (None, slice(None, None, None)),
        (slice(1, 2), slice(1, 2)),
        ('-1', slice(-1, None, None)),
        ('42', slice(42, 43)),
        ('-42', slice(-42, -41)),
        ('1:2:3', slice(1, 2, 3)),
        ('1::3', slice(1, None, 3)),
        (':', slice(None, None, None)),
        ('1:', slice(1, None, None)),
        ('[1:2:3]', slice(1, 2, 3)),
        ('(1:2:3)', slice(1, 2, 3)),
        ((4, 8, 10), slice(4, 8, 10)),
        ([10, 20], slice(10, 20))
        ])
def test_ensure_slice(inp, exp):
    obs = ensure_slice(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ((range(50), slice(25, 40)),
     list(i for i in range(25, 40))),

    (([1,2,3,4,5,6,7,8,9,10], [slice(1,4), slice(6, None)]),
     [2, 3, 4, 7, 8, 9, 10]),

    (([1,2,3,4,5], [slice(-2, None), slice(-5, -3)]),
     [4, 5, 1, 2]),
])
def test_get_portions(inp, exp):
    obs = get_portions(*inp)
    assert list(obs) == exp


@pytest.mark.parametrize('inp', [
    '42.3',
    '3:asd5:1',
    'test' ,
    '6.53:100:5',
    '4:-',
    '2:15-:3',
    '50:-:666',
    object(),
    [1,5,3,4],
    ('foo')
])
def test_ensure_slice_invalid(inp):
    with pytest.raises(ValueError):
        obs = ensure_slice(inp)


@pytest.mark.parametrize('inp, exp', [
    ('42', True),
    ('42.0', False),
    (42, False),
    ([42], False),
    ([], False),
    (None, False),
    ('', False),
    (False, False),
    (True, False),
])
def test_is_int_as_str(inp, exp):
    obs = is_int_as_str(inp)
    assert exp == obs
@pytest.mark.parametrize('inp, exp', [
    ('20', False),
    ('20%', False),
    ((20, 'c'), False),
    ((20.0, 'm'), False),
    ((20.0, 'c'), True),
    ((20.0, '%'), True),
])
def test_is_dynamic_cwd_width(inp, exp):
    obs = is_dynamic_cwd_width(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (42, False),
    (None, False),
    ('42', False),
    ('-42', False),
    (slice(1,2,3), False),
    ([], False),
    (False, False),
    (True, False),
    ('1:2:3', True),
    ('1::3', True),
    ('1:', True),
    (':', True),
    ('[1:2:3]', True),
    ('(1:2:3)', True),
    ('r', False),
    ('r:11', False),
])
def test_is_slice_as_str(inp, exp):
    obs = is_slice_as_str(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('throwback.log', True),
    ('', True),
    (None, True),
    (True, False),
    (False, False),
    (42, False),
    ([1, 2, 3], False),
    ((1, 2), False),
    (("wrong", "parameter"), False),
    skip_if_on_windows(('/dev/null', True))
])
def test_is_logfile_opt(inp, exp):
    obs = is_logfile_opt(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
        (True, None),
        (False, None),
        (1, None),
        (None, None),
        ('throwback.log', 'throwback.log'),
        skip_if_on_windows(('/dev/null', '/dev/null')),
        skip_if_on_windows(('/dev/nonexistent_dev', None))
    ])
def test_to_logfile_opt(inp, exp):
    obs = to_logfile_opt(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    (None, ''),
    ('', ''),
    ('throwback.log', 'throwback.log'),
    ('/dev/null', '/dev/null')
])
def test_logfile_opt_to_str(inp, exp):
    obs = logfile_opt_to_str(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ('20', (20.0, 'c')),
    ('20%', (20.0, '%')),
    ((20, 'c'), (20.0, 'c')),
    ((20, '%'), (20.0, '%')),
    ((20.0, 'c'), (20.0, 'c')),
    ((20.0, '%'), (20.0, '%')),
    ('inf', (float('inf'), 'c')),
])
def test_to_dynamic_cwd_tuple(inp, exp):
    obs = to_dynamic_cwd_tuple(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ((20.0, 'c'), '20.0'),
    ((20.0, '%'), '20.0%'),
    ((float('inf'), 'c'), 'inf'),
])
def test_dynamic_cwd_tuple_to_str(inp, exp):
    obs = dynamic_cwd_tuple_to_str(inp)
    assert exp == obs


@pytest.mark.parametrize('st, esc', [
    ('', ''),
    ('foo', 'foo'),
    ('foo&bar', 'foo^&bar'),
    ('foo$?-/_"\\', 'foo$?-/_^"\\'),
    ('^&<>|', '^^^&^<^>^|'),
    ('this /?', 'this /.')
])
def test_escape_windows_cmd_string(st, esc):
    obs = escape_windows_cmd_string(st)
    assert esc == obs


@pytest.mark.parametrize('st, esc, forced', [
    ('', '""', None),
    ('foo', 'foo', '"foo"'),
    (r'arg1 "hallo, "world""  "\some\path with\spaces")',
     r'"arg1 \"hallo, \"world\"\"  \"\some\path with\spaces\")"', None),
    (r'"argument"2" argument3 argument4',
     r'"\"argument\"2\" argument3 argument4"', None),
    (r'"\foo\bar bar\foo\" arg', r'"\"\foo\bar bar\foo\\\" arg"', None),
    (r'\\machine\dir\file.bat', r'\\machine\dir\file.bat', r'"\\machine\dir\file.bat"'),
    (r'"\\machine\dir space\file.bat"', r'"\"\\machine\dir space\file.bat\""', None)
])
def test_argvquote(st, esc, forced):
    obs = argvquote(st)
    assert esc == obs

    if forced is None:
        forced = esc
    obs = argvquote(st, force=True)
    assert forced == obs


@pytest.mark.parametrize('inp', ['no string here', ''])
def test_partial_string_none(inp):
    assert check_for_partial_string(inp) == (None, None, None)


@pytest.mark.parametrize('leaders', [
    (('', 0), ('not empty', 9)),
    (('not empty', 9), ('', 0))
])
@pytest.mark.parametrize('prefix', ['b', 'rb', 'r' ])
@pytest.mark.parametrize('quote', ['"', '"""'])
def test_partial_string(leaders, prefix, quote):
    (l, l_len), (f, f_len) = leaders
    s = prefix + quote
    t = s + 'test string' + quote
    t_len = len(t)
    # single string
    test_string = l + t + f
    obs = check_for_partial_string(test_string)
    exp = l_len, l_len + t_len, s
    assert obs == exp
    # single partial
    test_string = l + f + s + 'test string'
    obs = check_for_partial_string(test_string)
    exp = l_len + f_len, None, s
    assert obs == exp
    # two strings
    test_string = l + t + f + l + t + f
    obs = check_for_partial_string(test_string)
    exp = (l_len + t_len + f_len + l_len), (l_len + t_len + f_len + l_len + t_len), s
    assert obs == exp
    # one string, one partial
    test_string = l + t + f + l + s + 'test string'
    obs = check_for_partial_string(test_string)
    exp = l_len + t_len + f_len + l_len , None, s
    assert obs == exp


def test_executables_in(xonsh_builtins):
    expected = set()
    types = ('file', 'directory', 'brokensymlink')
    if ON_WINDOWS:
        # Don't test symlinks on windows since it requires admin
        types = ('file', 'directory')
    executables = (True, False)
    with TemporaryDirectory() as test_path:
        for _type in types:
            for executable in executables:
                fname = '%s_%s' % (_type, executable)
                if _type == 'none':
                    continue
                if _type == 'file' and executable:
                    ext = '.exe' if ON_WINDOWS else ''
                    expected.add(fname + ext)
                else:
                    ext = ''
                path = os.path.join(test_path, fname + ext)
                if _type == 'file':
                    with open(path, 'w') as f:
                        f.write(fname)
                elif _type == 'directory':
                    os.mkdir(path)
                elif _type == 'brokensymlink':
                    tmp_path = os.path.join(test_path, 'i_wont_exist')
                    with open(tmp_path, 'w') as f:
                        f.write('deleteme')
                        os.symlink(tmp_path, path)
                    os.remove(tmp_path)
                if executable and not _type == 'brokensymlink':
                    os.chmod(path, stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR)
            if ON_WINDOWS:
                xonsh_builtins.__xonsh_env__ = PATHEXT_ENV
                result = set(executables_in(test_path))
            else:
                result = set(executables_in(test_path))
    assert expected == result


@pytest.mark.parametrize('inp, exp', [
    ('yo', '[Yy][Oo]'),
    ('[a-f]123e', '[a-f]123[Ee]'),
    ('${HOME}/yo', '${HOME}/[Yy][Oo]'),
    ('./yo/mom', './[Yy][Oo]/[Mm][Oo][Mm]'),
    ('Eßen', '[Ee][Ss]?[Ssß][Ee][Nn]'),
])
def test_expand_case_matching(inp, exp):
    obs = expand_case_matching(inp)
    assert exp == obs


@pytest.mark.parametrize('inp, exp', [
    ("foo", "foo"),
    ("$foo $bar", "bar $bar"),
    ("$foobar", "$foobar"),
    ("$foo $spam", "bar eggs"),
    ("$an_int$spam$a_bool", "42eggsTrue"),
    ("bar$foo$spam$foo $an_int $none", "barbareggsbar 42 None"),
    ("$foo/bar", "bar/bar"),
    ("${'foo'} $spam", "bar eggs"),
    ("${'foo'} ${'a_bool'}", "bar True"),
    ("${'foo'}bar", "barbar"),
    ("${'foo'}/bar", "bar/bar"),
    ("${\"foo\'}", "${\"foo\'}"),
    ("$?bar", "$?bar"),
    ("$foo}bar", "bar}bar"),
    ("${'foo", "${'foo"),
    (b"foo", "foo"),
    (b"$foo bar", "bar bar"),
    (b"${'foo'}bar", "barbar"),
])
def test_expandvars(inp, exp, xonsh_builtins):
    """Tweaked for xonsh cases from CPython `test_genericpath.py`"""
    env = Env({'foo':'bar', 'spam': 'eggs', 'a_bool': True, 'an_int': 42, 'none': None})
    xonsh_builtins.__xonsh_env__ = env
    assert expandvars(inp) == exp


@pytest.mark.parametrize('inp, fmt, exp',[
    (572392800.0, None, 572392800.0),
    ('42.1459', None, 42.1459),
    (dt.datetime(2016, 8, 2, 13, 24), None, dt.datetime(2016, 8, 2, 13, 24).timestamp()),
    ('2016-8-10 16:14', None, dt.datetime(2016, 8, 10, 16, 14).timestamp()),
    ('2016/8/10 16:14:40', '%Y/%m/%d %H:%M:%S', dt.datetime(2016, 8, 10, 16, 14, 40).timestamp()),
    ])
def test_ensure_timestamp(inp, fmt, exp, xonsh_builtins):
    xonsh_builtins.__xonsh_env__['XONSH_DATETIME_FORMAT'] = '%Y-%m-%d %H:%M'
    obs = ensure_timestamp(inp, fmt)
    assert exp == obs


@pytest.mark.parametrize('expand_user', [True, False])
@pytest.mark.parametrize('inp, expand_env_vars, exp_end', [
    ('~/test.txt', True, '/test.txt'),
    ('~/$foo', True, '/bar'),
    ('~/test/$a_bool', True, '/test/True'),
    ('~/test/$an_int', True, '/test/42'),
    ('~/test/$none', True, '/test/None'),
    ('~/$foo', False, '/$foo')
])
def test_expand_path(expand_user, inp, expand_env_vars, exp_end, xonsh_builtins):
    if os.sep != '/':
        inp = inp.replace('/', os.sep)
        exp_end = exp_end.replace('/', os.sep)

    env = Env({'foo':'bar', 'a_bool': True, 'an_int': 42, 'none': None})
    env['EXPAND_ENV_VARS'] = expand_env_vars
    xonsh_builtins.__xonsh_env__ = env

    path = expand_path(inp, expand_user=expand_user)

    if expand_user:
        home_path = os.path.expanduser('~')
        assert path == home_path + exp_end
    else:
        assert path == '~' + exp_end


def test_swap_values():
    orig = {'x': 1}
    updates = {'x': 42, 'y': 43}
    with swap_values(orig, updates):
        assert orig['x'] == 42
        assert orig['y'] == 43
    assert orig['x'] == 1
    assert 'y' not in orig



# -*- coding: utf-8 -*-
"""Testing xonsh json hooks"""
import json

import pytest

from xonsh.tools import EnvPath
from xonsh.jsonutils import serialize_xonsh_json



@pytest.mark.parametrize('inp', [
    42, "yo", ["hello"], {"x": 65},
    EnvPath(["wakka", "jawaka"]),
    ["y", EnvPath(["wakka", "jawaka"])],
    {"z": EnvPath(["wakka", "jawaka"])},
])
def test_serialize_xonsh_json_roundtrip(inp):
    s = json.dumps(inp, default=serialize_xonsh_json)
    obs = json.loads(s)
    assert inp == obs


# -*- coding: utf-8 -*-
"""Tests lazy json functionality."""
from __future__ import unicode_literals, print_function
from io import StringIO

from xonsh.lazyjson import index, ljdump, LazyJSON, LJNode

def test_index_int():
    exp = {'offsets': 0, 'sizes': 2}
    s, obs = index(42)
    assert exp == obs

def test_index_str():
    exp = {'offsets': 0, 'sizes': 7}
    s, obs = index('wakka')
    assert exp == obs

def test_index_list_ints():
    exp = {'offsets': [1, 4, 0], 'sizes': [1, 2, 8]}
    s, obs = index([1, 42])
    assert exp == obs

def test_index_list_str():
    exp = {'offsets': [1, 10, 0], 'sizes': [7, 8, 20]}
    s, obs = index(['wakka', 'jawaka'])
    assert exp == obs

def test_index_list_str_int():
    exp = {'offsets': [1, 10, 0], 'sizes': [7, 2, 14]}
    s, obs = index(['wakka', 42])
    assert exp == obs

def test_index_list_int_str():
    exp = {'offsets': [1, 5, 14, 0], 'sizes': [2, 7, 8, 24]}
    s, obs = index([42, 'wakka', 'jawaka'])
    assert exp == obs

def test_index_dict_int():
    exp = {'offsets': {'wakka': 10, '__total__': 0},
           'sizes': {'wakka': 2, '__total__': 14}}
    s, obs = index({'wakka': 42})
    assert exp == obs

def test_index_dict_str():
    exp = {'offsets': {'wakka': 10, '__total__': 0},
           'sizes': {'wakka': 8, '__total__': 20}}
    s, obs = index({'wakka': 'jawaka'})
    assert exp == obs

def test_index_dict_dict_int():
    exp = {'offsets': {'wakka': {'jawaka': 21, '__total__': 10},
                      '__total__': 0,
                      },
           'sizes': {'wakka': {'jawaka': 2, '__total__': 15},
                    '__total__': 27}
           }
    s, obs = index({'wakka': {'jawaka': 42}})
    assert exp == obs

def test_lazy_load_index():
    f = StringIO()
    ljdump({'wakka': 42}, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert {'wakka': 10, '__total__': 0} == lj.offsets
    assert {'wakka': 2, '__total__': 14} == lj.sizes

def test_lazy_int():
    f = StringIO()
    ljdump(42, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 42 == lj.load()

def test_lazy_str():
    f = StringIO()
    ljdump('wakka', f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 'wakka' == lj.load()

def test_lazy_list_empty():
    x = []
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 0 == len(lj)
    assert x == lj.load()

def test_lazy_list_ints():
    x = [0, 1, 6, 28, 496, 8128]
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 28 == lj[3]
    assert x[:2:-2] == lj[:2:-2]
    assert x == [_ for _ in lj]
    assert x == lj.load()

def test_lazy_list_str():
    x = ['I', 'have', 'seen', 'the', 'wind', 'blow']
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 'the' == lj[3]
    assert x[:2:-2] == lj[:2:-2]
    assert x == [_ for _ in lj]
    assert x == lj.load()

def test_lazy_list_list_ints():
    x = [[0, 1], [6, 28], [496, 8128]]
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert isinstance(lj[1], LJNode)
    assert 28 == lj[1][1]
    assert [6 == 28], lj[1].load()
    assert x == lj.load()

def test_lazy_dict_empty():
    x = {}
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert 0 == len(lj)
    assert x == lj.load()

def test_lazy_dict():
    f = StringIO()
    ljdump({'wakka': 42}, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert ['wakka'] == list(lj.keys())
    assert 42 == lj['wakka']
    assert 1 == len(lj)
    assert {'wakka': 42} == lj.load()

def test_lazy_dict_dict_int():
    x = {'wakka': {'jawaka': 42}}
    f = StringIO()
    ljdump(x, f)
    f.seek(0)
    lj = LazyJSON(f)
    assert ['wakka'] == list(lj.keys())
    assert isinstance(lj['wakka'], LJNode)
    assert 42 == lj['wakka']['jawaka']
    assert 1 == len(lj)
    assert x == lj.load()



"""Implements a cat command for xonsh."""
import os

from xonsh.xoreutils.util import arg_handler


def _cat_single_file(opts, fname, stdin, out, err, line_count=1):
    if fname == '-':
        f = stdin
    elif os.path.isdir(fname):
        print("cat: {}: Is a directory.".format(fname), file=err)
        return True, line_count
    elif not os.path.exists(fname):
        print("cat: No such file or directory: {}".format(fname), file=err)
        return True, line_count
    else:
        f = open(fname, 'rb')
    sep = os.linesep.encode()
    last_was_blank = False
    while True:
        _r = r = f.readline()
        if isinstance(_r, str):
            _r = r = _r.encode()
        if r == b'':
            break
        if r.endswith(sep):
            _r = _r[:-len(sep)]
        this_one_blank = _r == b''
        if last_was_blank and this_one_blank and opts['squeeze_blank']:
            continue
        last_was_blank = this_one_blank
        if (opts['number_all'] or
                (opts['number_nonblank'] and not this_one_blank)):
            start = ("%6d " % line_count).encode()
            _r = start + _r
            line_count += 1
        if opts['show_ends']:
            _r = _r + b'$'
        try:
            print(_r.decode('unicode_escape'), flush=True, file=out)
        except:
            pass
    return False, line_count


def cat(args, stdin, stdout, stderr):
    """A cat command for xonsh."""
    opts = _cat_parse_args(args)
    if opts is None:
        print(CAT_HELP_STR, file=stdout)
        return 0

    line_count = 1
    errors = False
    if len(args) == 0:
        args = ['-']
    for i in args:
        o = _cat_single_file(opts, i, stdin, stdout, stderr, line_count)
        if o is None:
            return -1
        _e, line_count = o
        errors = _e or errors

    return int(errors)


def _cat_parse_args(args):
    out = {'number_nonblank': False, 'number_all': False, 'squeeze_blank': False, 'show_ends': False}
    if '--help' in args:
        return

    arg_handler(args, out, '-b', 'number_nonblank', True, '--number-nonblank')
    arg_handler(args, out, '-n', 'number_all', True, '--number')
    arg_handler(args, out, '-E', 'show_ends', True, '--show-ends')
    arg_handler(args, out, '-s', 'squeeze_blank', True, '--squeeze-blank')
    arg_handler(args, out, '-T', 'show_tabs', True, '--show-tabs')

    return out


CAT_HELP_STR = """This version of cat was written in Python for the xonsh project: http://xon.sh
Based on cat from GNU coreutils: http://www.gnu.org/software/coreutils/

Usage: cat [OPTION]... [FILE]...
Concatenate FILE(s), or standard input, to standard output.

  -b, --number-nonblank    number nonempty output lines, overrides -n
  -E, --show-ends          display $ at end of each line
  -n, --number             number all output lines
  -s, --squeeze-blank      suppress repeated empty output lines
  -T, --show-tabs          display TAB characters as ^I
  -u                       (ignored)
      --help     display this help and exit

With no FILE, or when FILE is -, read standard input.

Examples:
  cat f - g  Output f's contents, then standard input, then g's contents.
  cat        Copy standard input to standard output."""

# NOT IMPLEMENTED:
#  -A, --show-all           equivalent to -vET
#  -e                       equivalent to -vE
#  -t                       equivalent to -vT
#  -v, --show-nonprinting   use ^ and M- notation, except for LFD and TAB
#      --version  output version information and exit"""



# -*- coding: utf-8 -*-
"""Test XonshLexer for pygments"""

import os
import builtins

import pytest
from pygments.token import (Keyword, Name, String, Error, Number,
                            Operator, Punctuation, Text)
from tools import skip_if_on_windows

from xonsh.platform import ON_WINDOWS
from xonsh.built_ins import load_builtins, unload_builtins
from xonsh.pyghooks import XonshLexer


@pytest.yield_fixture(autouse=True)
def load_command_cache():
    load_builtins()
    if ON_WINDOWS:
        for key in ('cd', 'bash'):
            builtins.aliases[key] = lambda *args, **kwargs: None
    yield
    unload_builtins()


def check_token(code, tokens):
    """Make sure that all tokens appears in code in order"""
    lx = XonshLexer()
    tks = list(lx.get_tokens(code))

    for tk in tokens:
        while tks:
            if tk == tks[0]:
                break
            tks = tks[1:]
        else:
            msg = "Token {!r} missing: {!r}".format(tk,
                                                    list(lx.get_tokens(code)))
            pytest.fail(msg)
            break


@skip_if_on_windows
def test_ls():
    check_token('ls -al', [(Name.Builtin, 'ls')])


@skip_if_on_windows
def test_bin_ls():
    check_token('/bin/ls -al', [(Name.Builtin, '/bin/ls')])


def test_py_print():
    check_token('print("hello")', [(Keyword, 'print'),
                                   (String.Double, 'hello')])


def test_invalid_cmd():
    check_token('non-existance-cmd -al', [(Name, 'non')])  # parse as python
    check_token('![non-existance-cmd -al]',
                [(Error, 'non-existance-cmd')])  # parse as error
    check_token('for i in range(10):', [(Keyword, 'for')])  # as py keyword
    check_token('(1, )', [(Punctuation, '('),
                          (Number.Integer, '1')])


def test_multi_cmd():
    check_token('cd && cd', [(Name.Builtin, 'cd'),
                             (Operator, '&&'),
                             (Name.Builtin, 'cd')])
    check_token('cd || non-existance-cmd', [(Name.Builtin, 'cd'),
                                            (Operator, '||'),
                                            (Error, 'non-existance-cmd')
                                            ])


def test_nested():
    check_token('echo @("hello")', [(Name.Builtin, 'echo'),
                                    (Keyword, '@'),
                                    (Punctuation, '('),
                                    (String.Double, 'hello'),
                                    (Punctuation, ')')])
    check_token('print($(cd))', [(Keyword, 'print'),
                                 (Punctuation, '('),
                                 (Keyword, '$'),
                                 (Punctuation, '('),
                                 (Name.Builtin, 'cd'),
                                 (Punctuation, ')'),
                                 (Punctuation, ')')])
    check_token(r'print(![echo "])\""])', [(Keyword, 'print'),
                                           (Keyword, '!'),
                                           (Punctuation, '['),
                                           (Name.Builtin, 'echo'),
                                           (String.Double, r'"])\""'),
                                           (Punctuation, ']')])


def test_path(tmpdir):
    test_dir = str(tmpdir.mkdir('xonsh-test-highlight-path'))
    check_token('cd {}'.format(test_dir), [(Name.Builtin, 'cd'),
                                           (Name.Constant, test_dir)])
    check_token('cd {}-xxx'.format(test_dir), [(Name.Builtin, 'cd'),
                                               (Text,
                                                '{}-xxx'.format(test_dir))
                                               ])
    check_token('cd X={}'.format(test_dir), [(Name.Constant, test_dir)])

    with builtins.__xonsh_env__.swap(AUTO_CD=True):
        check_token(test_dir, [(Name.Constant, test_dir)])


def test_subproc_args():
    check_token('cd 192.168.0.1', [(Text, '192.168.0.1')])


def test_backtick():
    check_token(r'echo g`.*\w+`', [(String.Affix, 'g'),
                                   (String.Backtick, '`'),
                                   (String.Regex, '.'),
                                   (String.Regex, '*'),
                                   (String.Escape, r'\w'),
                                   ])


def test_macro():
    check_token(r'g!(42, *, 65)', [(Name, 'g'),
                                   (Keyword, '!'),
                                   (Punctuation, '('),
                                   (Number.Integer, '42')])
    check_token(r'echo! hello world', [(Name.Builtin, 'echo'),
                                       (Keyword, '!'),
                                       (String, "hello world")])
    check_token(r'bash -c ! export var=42; echo $var',
                [(Name.Builtin, 'bash'),
                 (Text, '-c'),
                 (Keyword, '!'),
                 (String, 'export var=42; echo $var'),
                 ])



# -*- coding: utf-8 -*-
"""Informative git status prompt formatter"""

import builtins
import collections
import os
import subprocess

import xonsh.lazyasd as xl


GitStatus = collections.namedtuple('GitStatus',
                                   ['branch', 'num_ahead', 'num_behind',
                                    'untracked', 'changed', 'conflicts',
                                    'staged', 'stashed', 'operations'])


def _check_output(*args, **kwargs):
    kwargs.update(dict(env=builtins.__xonsh_env__.detype(),
                       stderr=subprocess.DEVNULL,
                       timeout=builtins.__xonsh_env__['VC_BRANCH_TIMEOUT'],
                       universal_newlines=True
                       ))
    return subprocess.check_output(*args, **kwargs)


@xl.lazyobject
def _DEFS():
    DEFS = {
        'HASH': ':',
        'BRANCH': '{CYAN}',
        'OPERATION': '{CYAN}',
        'STAGED': '{RED}●',
        'CONFLICTS': '{RED}×',
        'CHANGED': '{BLUE}+',
        'UNTRACKED': '…',
        'STASHED': '⚑',
        'CLEAN': '{BOLD_GREEN}✓',
        'AHEAD': '↑·',
        'BEHIND': '↓·',
    }
    return DEFS


def _get_def(key):
    def_ = builtins.__xonsh_env__.get('XONSH_GITSTATUS_' + key)
    return def_ if def_ is not None else _DEFS[key]


def _get_tag_or_hash():
    tag = _check_output(['git', 'describe', '--exact-match']).strip()
    if tag:
        return tag
    hash_ = _check_output(['git', 'rev-parse', '--short', 'HEAD']).strip()
    return _get_def('HASH') + hash_


def _get_stash(gitdir):
    try:
        with open(os.path.join(gitdir, 'logs/refs/stash')) as f:
            return sum(1 for _ in f)
    except IOError:
        return 0


def _gitoperation(gitdir):
    files = (
             ('rebase-merge', 'REBASE'),
             ('rebase-apply', 'AM/REBASE'),
             ('MERGE_HEAD', 'MERGING'),
             ('CHERRY_PICK_HEAD', 'CHERRY-PICKING'),
             ('REVERT_HEAD', 'REVERTING'),
             ('BISECT_LOG', 'BISECTING'),
             )
    return [f[1] for f in files
            if os.path.exists(os.path.join(gitdir, f[0]))]


def gitstatus():
    """Return namedtuple with fields:
    branch name, number of ahead commit, number of behind commit,
    untracked number, changed number, conflicts number,
    staged number, stashed number, operation."""
    status = _check_output(['git', 'status', '--porcelain', '--branch'])
    branch = ''
    num_ahead, num_behind = 0, 0
    untracked, changed, conflicts, staged = 0, 0, 0, 0
    for line in status.splitlines():
        if line.startswith('##'):
            line = line[2:].strip()
            if 'Initial commit on' in line:
                branch = line.split()[-1]
            elif 'no branch' in line:
                branch = _get_tag_or_hash()
            elif '...' not in line:
                branch = line
            else:
                branch, rest = line.split('...')
                if ' ' in rest:
                    divergence = rest.split(' ', 1)[-1]
                    divergence = divergence.strip('[]')
                    for div in divergence.split(', '):
                        if 'ahead' in div:
                            num_ahead = int(div[len('ahead '):].strip())
                        elif 'behind' in div:
                            num_behind = int(div[len('behind '):].strip())
        elif line.startswith('??'):
            untracked += 1
        else:
            if len(line) > 1 and line[1] == 'M':
                changed += 1

            if len(line) > 0 and line[0] == 'U':
                conflicts += 1
            elif len(line) > 0 and line[0] != ' ':
                staged += 1

    gitdir = _check_output(['git', 'rev-parse', '--git-dir']).strip()
    stashed = _get_stash(gitdir)
    operations = _gitoperation(gitdir)

    return GitStatus(branch, num_ahead, num_behind,
                     untracked, changed, conflicts, staged, stashed,
                     operations)


def gitstatus_prompt():
    """Return str `BRANCH|OPERATOR|numbers`"""
    try:
        s = gitstatus()
    except subprocess.SubprocessError:
        return None

    ret = _get_def('BRANCH') + s.branch
    if s.num_ahead > 0:
        ret += _get_def('AHEAD') + str(s.num_ahead)
    if s.num_behind > 0:
        ret += _get_def('BEHIND') + str(s.num_behind)
    if s.operations:
        ret += _get_def('OPERATION') + '|' + '|'.join(s.operations)
    ret += '|'
    if s.staged > 0:
        ret += _get_def('STAGED') + str(s.staged) + '{NO_COLOR}'
    if s.conflicts > 0:
        ret += _get_def('CONFLICTS') + str(s.conflicts) + '{NO_COLOR}'
    if s.changed > 0:
        ret += _get_def('CHANGED') + str(s.changed) + '{NO_COLOR}'
    if s.untracked > 0:
        ret += _get_def('UNTRACKED') + str(s.untracked) + '{NO_COLOR}'
    if s.stashed > 0:
        ret += _get_def('STASHED') + str(s.stashed) + '{NO_COLOR}'
    if s.staged + s.conflicts + s.changed + s.untracked + s.stashed == 0:
        ret += _get_def('CLEAN') + '{NO_COLOR}'
    ret += '{NO_COLOR}'

    return ret



"""Xonsh color styling tools that simulate pygments, when it is unavailable."""
import builtins
import string

from xonsh.platform import HAS_PYGMENTS
from xonsh.lazyasd import LazyObject
from xonsh.color_tools import RE_BACKGROUND


class _TokenType(tuple):
    """
    Forked from the pygments project
    https://bitbucket.org/birkenfeld/pygments-main
    Copyright (c) 2006-2017 by the respective authors, All rights reserved.
    See https://bitbucket.org/birkenfeld/pygments-main/raw/05818a4ef9891d9ac22c851f7b3ea4b4fce460ab/AUTHORS
    """
    parent = None

    def split(self):
        buf = []
        node = self
        while node is not None:
            buf.append(node)
            node = node.parent
        buf.reverse()
        return buf

    def __init__(self, *args):
        # no need to call super.__init__
        self.subtypes = set()

    def __contains__(self, val):
        return self is val or (
            type(val) is self.__class__ and
            val[:len(self)] == self
        )

    def __getattr__(self, val):
        if not val or not val[0].isupper():
            return tuple.__getattribute__(self, val)
        new = _TokenType(self + (val,))
        setattr(self, val, new)
        self.subtypes.add(new)
        new.parent = self
        return new

    def __repr__(self):
        return 'Token' + (self and '.' or '') + '.'.join(self)

    def __copy__(self):
        # These instances are supposed to be singletons
        return self

    def __deepcopy__(self, memo):
        # These instances are supposed to be singletons
        return self


Token = _TokenType()
Color = Token.Color


def partial_color_tokenize(template):
    """Tokenizes a template string containing colors. Will return a list
    of tuples mapping the token to the string which has that color.
    These sub-strings maybe templates themselves.
    """
    if HAS_PYGMENTS and hasattr(builtins, '__xonsh_shell__'):
        styles = __xonsh_shell__.shell.styler.styles
    elif hasattr(builtins, '__xonsh_shell__'):
        styles = DEFAULT_STYLE_DICT
    else:
        styles = None
    color = Color.NO_COLOR
    try:
        toks, color = _partial_color_tokenize_main(template, styles)
    except Exception:
        toks = [(Color.NO_COLOR, template)]
    if styles is not None:
        styles[color]  # ensure color is available
    return toks


def _partial_color_tokenize_main(template, styles):
    formatter = string.Formatter()
    bopen = '{'
    bclose = '}'
    colon = ':'
    expl = '!'
    color = Color.NO_COLOR
    fg = bg = None
    value = ''
    toks = []
    for literal, field, spec, conv in formatter.parse(template):
        if field is None:
            value += literal
        elif field in KNOWN_COLORS or '#' in field:
            value += literal
            next_color, fg, bg = color_by_name(field, fg, bg)
            if next_color is not color:
                if len(value) > 0:
                    toks.append((color, value))
                    if styles is not None:
                        styles[color]  # ensure color is available
                color = next_color
                value = ''
        elif field is not None:
            parts = [literal, bopen, field]
            if conv is not None and len(conv) > 0:
                parts.append(expl)
                parts.append(conv)
            if spec is not None and len(spec) > 0:
                parts.append(colon)
                parts.append(spec)
            parts.append(bclose)
            value += ''.join(parts)
        else:
            value += literal
    toks.append((color, value))
    return toks, color


def color_by_name(name, fg=None, bg=None):
    """Converts a color name to a color token, foreground name,
    and background name.  Will take into consideration current foreground
    and background colors, if provided.

    Parameters
    ----------
    name : str
        Color name.
    fg : str, optional
        Foreground color name.
    bg : str, optional
        Background color name.

    Returns
    -------
    tok : Token
        Pygments Token.Color subclass
    fg : str or None
        New computed foreground color name.
    bg : str or None
        New computed background color name.
    """
    name = name.upper()
    if name == 'NO_COLOR':
        return Color.NO_COLOR, None, None
    m = RE_BACKGROUND.search(name)
    if m is None:  # must be foreground color
        fg = norm_name(name)
    else:
        bg = norm_name(name)
    # assmble token
    if fg is None and bg is None:
        tokname = 'NO_COLOR'
    elif fg is None:
        tokname = bg
    elif bg is None:
        tokname = fg
    else:
        tokname = fg + '__' + bg
    tok = getattr(Color, tokname)
    return tok, fg, bg


def norm_name(name):
    """Normalizes a color name."""
    return name.replace('#', 'HEX').replace('BGHEX', 'BACKGROUND_HEX')

KNOWN_COLORS = LazyObject(lambda: frozenset([
    'BACKGROUND_BLACK',
    'BACKGROUND_BLUE',
    'BACKGROUND_CYAN',
    'BACKGROUND_GREEN',
    'BACKGROUND_INTENSE_BLACK',
    'BACKGROUND_INTENSE_BLUE',
    'BACKGROUND_INTENSE_CYAN',
    'BACKGROUND_INTENSE_GREEN',
    'BACKGROUND_INTENSE_PURPLE',
    'BACKGROUND_INTENSE_RED',
    'BACKGROUND_INTENSE_WHITE',
    'BACKGROUND_INTENSE_YELLOW',
    'BACKGROUND_PURPLE',
    'BACKGROUND_RED',
    'BACKGROUND_WHITE',
    'BACKGROUND_YELLOW',
    'BLACK',
    'BLUE',
    'BOLD_BLACK',
    'BOLD_BLUE',
    'BOLD_CYAN',
    'BOLD_GREEN',
    'BOLD_INTENSE_BLACK',
    'BOLD_INTENSE_BLUE',
    'BOLD_INTENSE_CYAN',
    'BOLD_INTENSE_GREEN',
    'BOLD_INTENSE_PURPLE',
    'BOLD_INTENSE_RED',
    'BOLD_INTENSE_WHITE',
    'BOLD_INTENSE_YELLOW',
    'BOLD_PURPLE',
    'BOLD_RED',
    'BOLD_UNDERLINE_BLACK',
    'BOLD_UNDERLINE_BLUE',
    'BOLD_UNDERLINE_CYAN',
    'BOLD_UNDERLINE_GREEN',
    'BOLD_UNDERLINE_INTENSE_BLACK',
    'BOLD_UNDERLINE_INTENSE_BLUE',
    'BOLD_UNDERLINE_INTENSE_CYAN',
    'BOLD_UNDERLINE_INTENSE_GREEN',
    'BOLD_UNDERLINE_INTENSE_PURPLE',
    'BOLD_UNDERLINE_INTENSE_RED',
    'BOLD_UNDERLINE_INTENSE_WHITE',
    'BOLD_UNDERLINE_INTENSE_YELLOW',
    'BOLD_UNDERLINE_PURPLE',
    'BOLD_UNDERLINE_RED',
    'BOLD_UNDERLINE_WHITE',
    'BOLD_UNDERLINE_YELLOW',
    'BOLD_WHITE',
    'BOLD_YELLOW',
    'CYAN',
    'GREEN',
    'INTENSE_BLACK',
    'INTENSE_BLUE',
    'INTENSE_CYAN',
    'INTENSE_GREEN',
    'INTENSE_PURPLE',
    'INTENSE_RED',
    'INTENSE_WHITE',
    'INTENSE_YELLOW',
    'NO_COLOR',
    'PURPLE',
    'RED',
    'UNDERLINE_BLACK',
    'UNDERLINE_BLUE',
    'UNDERLINE_CYAN',
    'UNDERLINE_GREEN',
    'UNDERLINE_INTENSE_BLACK',
    'UNDERLINE_INTENSE_BLUE',
    'UNDERLINE_INTENSE_CYAN',
    'UNDERLINE_INTENSE_GREEN',
    'UNDERLINE_INTENSE_PURPLE',
    'UNDERLINE_INTENSE_RED',
    'UNDERLINE_INTENSE_WHITE',
    'UNDERLINE_INTENSE_YELLOW',
    'UNDERLINE_PURPLE',
    'UNDERLINE_RED',
    'UNDERLINE_WHITE',
    'UNDERLINE_YELLOW',
    'WHITE',
    'YELLOW',
    ]), globals(), 'KNOWN_COLORS')

DEFAULT_STYLE_DICT = LazyObject(lambda: {
    Token: '',
    Token.Aborted: '#ansidarkgray',
    Token.AutoSuggestion: '#ansidarkgray',
    Token.Color.BACKGROUND_BLACK: 'bg:#ansiblack',
    Token.Color.BACKGROUND_BLUE: 'bg:#ansidarkblue',
    Token.Color.BACKGROUND_CYAN: 'bg:#ansiteal',
    Token.Color.BACKGROUND_GREEN: 'bg:#ansidarkgreen',
    Token.Color.BACKGROUND_INTENSE_BLACK: 'bg:#ansidarkgray',
    Token.Color.BACKGROUND_INTENSE_BLUE: 'bg:#ansiblue',
    Token.Color.BACKGROUND_INTENSE_CYAN: 'bg:#ansiturquoise',
    Token.Color.BACKGROUND_INTENSE_GREEN: 'bg:#ansigreen',
    Token.Color.BACKGROUND_INTENSE_PURPLE: 'bg:#ansifuchsia',
    Token.Color.BACKGROUND_INTENSE_RED: 'bg:#ansired',
    Token.Color.BACKGROUND_INTENSE_WHITE: 'bg:#ansiwhite',
    Token.Color.BACKGROUND_INTENSE_YELLOW: 'bg:#ansiyellow',
    Token.Color.BACKGROUND_PURPLE: 'bg:#ansipurple',
    Token.Color.BACKGROUND_RED: 'bg:#ansidarkred',
    Token.Color.BACKGROUND_WHITE: 'bg:#ansilightgray',
    Token.Color.BACKGROUND_YELLOW: 'bg:#ansibrown',
    Token.Color.BLACK: '#ansiblack',
    Token.Color.BLUE: '#ansidarkblue',
    Token.Color.BOLD_BLACK: 'bold #ansiblack',
    Token.Color.BOLD_BLUE: 'bold #ansidarkblue',
    Token.Color.BOLD_CYAN: 'bold #ansiteal',
    Token.Color.BOLD_GREEN: 'bold #ansidarkgreen',
    Token.Color.BOLD_INTENSE_BLACK: 'bold #ansidarkgray',
    Token.Color.BOLD_INTENSE_BLUE: 'bold #ansiblue',
    Token.Color.BOLD_INTENSE_CYAN: 'bold #ansiturquoise',
    Token.Color.BOLD_INTENSE_GREEN: 'bold #ansigreen',
    Token.Color.BOLD_INTENSE_PURPLE: 'bold #ansifuchsia',
    Token.Color.BOLD_INTENSE_RED: 'bold #ansired',
    Token.Color.BOLD_INTENSE_WHITE: 'bold #ansiwhite',
    Token.Color.BOLD_INTENSE_YELLOW: 'bold #ansiyellow',
    Token.Color.BOLD_PURPLE: 'bold #ansipurple',
    Token.Color.BOLD_RED: 'bold #ansidarkred',
    Token.Color.BOLD_UNDERLINE_BLACK: 'bold underline #ansiblack',
    Token.Color.BOLD_UNDERLINE_BLUE: 'bold underline #ansidarkblue',
    Token.Color.BOLD_UNDERLINE_CYAN: 'bold underline #ansiteal',
    Token.Color.BOLD_UNDERLINE_GREEN: 'bold underline #ansidarkgreen',
    Token.Color.BOLD_UNDERLINE_INTENSE_BLACK: 'bold underline #ansidarkgray',
    Token.Color.BOLD_UNDERLINE_INTENSE_BLUE: 'bold underline #ansiblue',
    Token.Color.BOLD_UNDERLINE_INTENSE_CYAN: 'bold underline #ansiturquoise',
    Token.Color.BOLD_UNDERLINE_INTENSE_GREEN: 'bold underline #ansigreen',
    Token.Color.BOLD_UNDERLINE_INTENSE_PURPLE: 'bold underline #ansifuchsia',
    Token.Color.BOLD_UNDERLINE_INTENSE_RED: 'bold underline #ansired',
    Token.Color.BOLD_UNDERLINE_INTENSE_WHITE: 'bold underline #ansiwhite',
    Token.Color.BOLD_UNDERLINE_INTENSE_YELLOW: 'bold underline #ansiyellow',
    Token.Color.BOLD_UNDERLINE_PURPLE: 'bold underline #ansipurple',
    Token.Color.BOLD_UNDERLINE_RED: 'bold underline #ansidarkred',
    Token.Color.BOLD_UNDERLINE_WHITE: 'bold underline #ansilightgray',
    Token.Color.BOLD_UNDERLINE_YELLOW: 'bold underline #ansibrown',
    Token.Color.BOLD_WHITE: 'bold #ansilightgray',
    Token.Color.BOLD_YELLOW: 'bold #ansibrown',
    Token.Color.CYAN: '#ansiteal',
    Token.Color.GREEN: '#ansidarkgreen',
    Token.Color.INTENSE_BLACK: '#ansidarkgray',
    Token.Color.INTENSE_BLUE: '#ansiblue',
    Token.Color.INTENSE_CYAN: '#ansiturquoise',
    Token.Color.INTENSE_GREEN: '#ansigreen',
    Token.Color.INTENSE_PURPLE: '#ansifuchsia',
    Token.Color.INTENSE_RED: '#ansired',
    Token.Color.INTENSE_WHITE: '#ansiwhite',
    Token.Color.INTENSE_YELLOW: '#ansiyellow',
    Token.Color.NO_COLOR: 'noinherit',
    Token.Color.PURPLE: '#ansipurple',
    Token.Color.RED: '#ansidarkred',
    Token.Color.UNDERLINE_BLACK: 'underline #ansiblack',
    Token.Color.UNDERLINE_BLUE: 'underline #ansidarkblue',
    Token.Color.UNDERLINE_CYAN: 'underline #ansiteal',
    Token.Color.UNDERLINE_GREEN: 'underline #ansidarkgreen',
    Token.Color.UNDERLINE_INTENSE_BLACK: 'underline #ansidarkgray',
    Token.Color.UNDERLINE_INTENSE_BLUE: 'underline #ansiblue',
    Token.Color.UNDERLINE_INTENSE_CYAN: 'underline #ansiturquoise',
    Token.Color.UNDERLINE_INTENSE_GREEN: 'underline #ansigreen',
    Token.Color.UNDERLINE_INTENSE_PURPLE: 'underline #ansifuchsia',
    Token.Color.UNDERLINE_INTENSE_RED: 'underline #ansired',
    Token.Color.UNDERLINE_INTENSE_WHITE: 'underline #ansiwhite',
    Token.Color.UNDERLINE_INTENSE_YELLOW: 'underline #ansiyellow',
    Token.Color.UNDERLINE_PURPLE: 'underline #ansipurple',
    Token.Color.UNDERLINE_RED: 'underline #ansidarkred',
    Token.Color.UNDERLINE_WHITE: 'underline #ansilightgray',
    Token.Color.UNDERLINE_YELLOW: 'underline #ansibrown',
    Token.Color.WHITE: '#ansilightgray',
    Token.Color.YELLOW: '#ansibrown',
    Token.Comment: 'underline #ansiteal',
    Token.Comment.Hashbang: '',
    Token.Comment.Multiline: '',
    Token.Comment.Preproc: 'underline #ansibrown',
    Token.Comment.PreprocFile: '',
    Token.Comment.Single: '',
    Token.Comment.Special: '',
    Token.Error: '#ansired',
    Token.Escape: '',
    Token.Generic: '',
    Token.Generic.Deleted: '#ansidarkred',
    Token.Generic.Emph: 'underline',
    Token.Generic.Error: 'bold #ansired',
    Token.Generic.Heading: 'bold #ansidarkblue',
    Token.Generic.Inserted: '#ansigreen',
    Token.Generic.Output: '#ansidarkblue',
    Token.Generic.Prompt: 'bold #ansidarkblue',
    Token.Generic.Strong: '',
    Token.Generic.Subheading: 'bold #ansipurple',
    Token.Generic.Traceback: '#ansidarkblue',
    Token.Keyword: 'bold #ansidarkgreen',
    Token.Keyword.Constant: '',
    Token.Keyword.Declaration: '',
    Token.Keyword.Namespace: '',
    Token.Keyword.Pseudo: 'nobold',
    Token.Keyword.Reserved: '',
    Token.Keyword.Type: 'nobold #ansidarkred',
    Token.Literal: '',
    Token.Literal.Date: '',
    Token.Literal.Number: '#ansidarkgray',
    Token.Literal.Number.Bin: '',
    Token.Literal.Number.Float: '',
    Token.Literal.Number.Hex: '',
    Token.Literal.Number.Integer: '',
    Token.Literal.Number.Integer.Long: '',
    Token.Literal.Number.Oct: '',
    Token.Literal.String: '#ansired',
    Token.Literal.String.Affix: '',
    Token.Literal.String.Backtick: '',
    Token.Literal.String.Char: '',
    Token.Literal.String.Delimiter: '',
    Token.Literal.String.Doc: 'underline',
    Token.Literal.String.Double: '',
    Token.Literal.String.Escape: 'bold #ansibrown',
    Token.Literal.String.Heredoc: '',
    Token.Literal.String.Interpol: 'bold #ansipurple',
    Token.Literal.String.Other: '#ansidarkgreen',
    Token.Literal.String.Regex: '#ansipurple',
    Token.Literal.String.Single: '',
    Token.Literal.String.Symbol: '#ansibrown',
    Token.Menu.Completions: 'bg:#ansilightgray #ansiblack',
    Token.Menu.Completions.Completion: '',
    Token.Menu.Completions.Completion.Current: 'bg:#ansidarkgray #ansiwhite',
    Token.Name: '',
    Token.Name.Attribute: '#ansiyellow',
    Token.Name.Builtin: '#ansidarkgreen',
    Token.Name.Builtin.Pseudo: '',
    Token.Name.Class: 'bold #ansiblue',
    Token.Name.Constant: '#ansidarkred',
    Token.Name.Decorator: '#ansifuchsia',
    Token.Name.Entity: 'bold #ansilightgray',
    Token.Name.Exception: 'bold #ansired',
    Token.Name.Function: '#ansiblue',
    Token.Name.Function.Magic: '',
    Token.Name.Label: '#ansiyellow',
    Token.Name.Namespace: 'bold #ansiblue',
    Token.Name.Other: '',
    Token.Name.Property: '',
    Token.Name.Tag: 'bold #ansidarkgreen',
    Token.Name.Variable: '#ansidarkblue',
    Token.Name.Variable.Class: '',
    Token.Name.Variable.Global: '',
    Token.Name.Variable.Instance: '',
    Token.Name.Variable.Magic: '',
    Token.Operator: '#ansidarkgray',
    Token.Operator.Word: 'bold #ansipurple',
    Token.Other: '',
    Token.Punctuation: '',
    Token.Scrollbar: 'bg:#ansidarkgray',
    Token.Scrollbar.Arrow: 'bg:#ansiblack #ansiwhite bold',
    Token.Scrollbar.Button: 'bg:#ansiblack',
    Token.Text: '',
    Token.Text.Whitespace: '#ansilightgray'},
                                globals(), 'DEFAULT_STYLE_DICT')



# -*- coding: utf-8 -*-
"""Implements the xonsh history backend."""
import collections
from xonsh.history.base import History


class DummyHistory(History):
    """A dummy implement of history backend."""
    def append(self, cmd):
        pass

    def items(self):
        yield {'inp': 'dummy in action', 'ts': 1464652800, 'ind': 0}

    def all_items(self):
        return self.items()

    def info(self):
        data = collections.OrderedDict()
        data['backend'] = 'dummy'
        data['sessionid'] = str(self.sessionid)
        return data



"""Xonsh hooks into bash completions."""
import builtins

import xonsh.platform as xp
from xonsh.completers.path import _quote_paths
from xonsh.completers.bash_completion import bash_completions


def complete_from_bash(prefix, line, begidx, endidx, ctx):
    """Completes based on results from BASH completion."""
    env = builtins.__xonsh_env__.detype()
    paths = builtins.__xonsh_env__.get('BASH_COMPLETIONS', ())
    command = xp.bash_command()
    return bash_completions(prefix, line, begidx, endidx, env=env, paths=paths,
                            command=command, quote_paths=_quote_paths)



# -*- coding: utf-8 -*-
"""Tests the xonsh history."""
# pylint: disable=protected-access
import os
import shlex

from xonsh.history.sqlite import SqliteHistory
from xonsh.history.main import history_main

import pytest


@pytest.yield_fixture
def hist():
    h = SqliteHistory(filename='xonsh-HISTORY-TEST.sqlite',
                      sessionid='SESSIONID', gc=False)
    yield h
    os.remove(h.filename)


def test_hist_append(hist, xonsh_builtins):
    """Verify appending to the history works."""
    xonsh_builtins.__xonsh_env__['HISTCONTROL'] = set()
    hf = hist.append({'inp': 'still alive', 'rtn': 1})
    assert hf is None
    items = list(hist.items())
    assert len(items) == 1
    assert 'still alive' == items[0]['inp']
    assert 1 == items[0]['rtn']
    hist.append({'inp': 'still alive', 'rtn': 0})
    items = list(hist.items())
    assert len(items) == 2
    assert 'still alive' == items[1]['inp']
    assert 0 == items[1]['rtn']
    assert list(hist.all_items()) == items


def test_hist_attrs(hist, xonsh_builtins):
    xonsh_builtins.__xonsh_env__['HISTCONTROL'] = set()
    hf = hist.append({'inp': 'ls foo', 'rtn': 1})
    assert hf is None
    assert 'ls foo' == hist.inps[0]
    assert 'ls foo' == hist.inps[-1]
    assert 1 == hist.rtns[0]
    assert 1 == hist.rtns[-1]
    assert None is hist.outs[-1]
    assert [1] == hist.rtns[:]
    hist.append({'inp': 'ls bar', 'rtn': 0})
    assert 'ls bar' == hist.inps[1]
    assert 'ls bar' == hist.inps[-1]
    assert 0 == hist.rtns[1]
    assert 0 == hist.rtns[-1]
    assert None is hist.outs[-1]
    assert [1, 0] == hist.rtns[:]
    assert len(hist.tss) == 2
    assert len(hist.tss[0]) == 2


CMDS = ['ls', 'cat hello kitty', 'abc', 'def', 'touch me', 'grep from me']


@pytest.mark.parametrize('inp, commands, offset', [
    ('', CMDS, (0, 1)),
    ('-r', list(reversed(CMDS)), (len(CMDS) - 1, -1)),
    ('0', CMDS[0:1], (0, 1)),
    ('1', CMDS[1:2], (1, 1)),
    ('-2', CMDS[-2:-1], (len(CMDS) - 2, 1)),
    ('1:3', CMDS[1:3], (1, 1)),
    ('1::2', CMDS[1::2], (1, 2)),
    ('-4:-2', CMDS[-4:-2], (len(CMDS) - 4, 1))
    ])
def test_show_cmd_numerate(inp, commands, offset, hist, xonsh_builtins, capsys):
    """Verify that CLI history commands work."""
    base_idx, step = offset
    xonsh_builtins.__xonsh_history__ = hist
    xonsh_builtins.__xonsh_env__['HISTCONTROL'] = set()
    for ts, cmd in enumerate(CMDS):  # populate the shell history
        hist.append({'inp': cmd, 'rtn': 0, 'ts': (ts + 1, ts + 1.5)})

    exp = ('{}: {}'.format(base_idx + idx * step, cmd)
           for idx, cmd in enumerate(list(commands)))
    exp = '\n'.join(exp)

    history_main(['show', '-n'] + shlex.split(inp))
    out, err = capsys.readouterr()
    assert out.rstrip() == exp


def test_histcontrol(hist, xonsh_builtins):
    """Test HISTCONTROL=ignoredups,ignoreerr"""

    xonsh_builtins.__xonsh_env__['HISTCONTROL'] = 'ignoredups,ignoreerr'
    assert len(hist) == 0

    # An error, items() remains empty
    hist.append({'inp': 'ls foo', 'rtn': 2})
    assert len(hist) == 0
    assert len(hist.inps) == 1
    assert len(hist.rtns) == 1
    assert 2 == hist.rtns[-1]

    # Success
    hist.append({'inp': 'ls foobazz', 'rtn': 0})
    assert len(hist) == 1
    assert len(hist.inps) == 2
    assert len(hist.rtns) == 2
    items = list(hist.items())
    assert 'ls foobazz' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 0 == hist.rtns[-1]

    # Error
    hist.append({'inp': 'ls foo', 'rtn': 2})
    assert len(hist) == 1
    items = list(hist.items())
    assert 'ls foobazz' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 2 == hist.rtns[-1]

    # File now exists, success
    hist.append({'inp': 'ls foo', 'rtn': 0})
    assert len(hist) == 2
    items = list(hist.items())
    assert 'ls foo' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 0 == hist.rtns[-1]

    # Success
    hist.append({'inp': 'ls', 'rtn': 0})
    assert len(hist) == 3
    items = list(hist.items())
    assert 'ls' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 0 == hist.rtns[-1]

    # Dup
    hist.append({'inp': 'ls', 'rtn': 0})
    assert len(hist) == 3

    # Success
    hist.append({'inp': '/bin/ls', 'rtn': 0})
    assert len(hist) == 4
    items = list(hist.items())
    assert '/bin/ls' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 0 == hist.rtns[-1]

    # Error
    hist.append({'inp': 'ls bazz', 'rtn': 1})
    assert len(hist) == 4
    items = list(hist.items())
    assert '/bin/ls' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert 'ls bazz' == hist.inps[-1]
    assert 1 == hist.rtns[-1]

    # Error
    hist.append({'inp': 'ls bazz', 'rtn': -1})
    assert len(hist) == 4
    items = list(hist.items())
    assert '/bin/ls' == items[-1]['inp']
    assert 0 == items[-1]['rtn']
    assert -1 == hist.rtns[-1]


@pytest.mark.parametrize('index, exp', [
    (-1, ('grep from me', 'out', 0, (5, 6))),
    (1, ('cat hello kitty', 'out', 0, (1, 2))),
    (slice(1, 3), [('cat hello kitty', 'out', 0, (1, 2)),
                   ('abc', 'out', 0, (2, 3))]),
])
def test_history_getitem(index, exp, hist, xonsh_builtins):
    xonsh_builtins.__xonsh_env__['HISTCONTROL'] = set()
    xonsh_builtins.__xonsh_env__['XONSH_STORE_STDOUT'] = True
    attrs = ('inp', 'out', 'rtn', 'ts')

    for ts, cmd in enumerate(CMDS):  # populate the shell history
        entry = {k: v for k, v in zip(attrs, [cmd, 'out', 0, (ts, ts + 1)])}
        hist.append(entry)

    entry = hist[index]
    if isinstance(entry, list):
        assert [(e.cmd, e.out, e.rtn, e.ts) for e in entry] == exp
    else:
        assert (entry.cmd, entry.out, entry.rtn, entry.ts) == exp



"""Provides completions for xonsh internal utilities"""

import xonsh.xontribs as xx
import xonsh.tools as xt


def complete_xonfig(prefix, line, start, end, ctx):
    """Completion for ``xonfig``"""
    args = line.split(' ')
    if len(args) == 0 or args[0] != 'xonfig':
        return None
    curix = args.index(prefix)
    if curix == 1:
        possible = {'info', 'wizard', 'styles', 'colors', '-h'}
    elif curix == 2 and args[1] == 'colors':
        possible = set(xt.color_style_names())
    else:
        raise StopIteration
    return {i for i in possible if i.startswith(prefix)}


def _list_installed_xontribs():
    meta = xx.xontrib_metadata()
    installed = []
    for md in meta['xontribs']:
        name = md['name']
        spec = xx.find_xontrib(name)
        if spec is not None:
            installed.append(spec.name.rsplit('.')[-1])

    return installed


def complete_xontrib(prefix, line, start, end, ctx):
    """Completion for ``xontrib``"""
    args = line.split(' ')
    if len(args) == 0 or args[0] != 'xontrib':
        return None
    curix = args.index(prefix)
    if curix == 1:
        possible = {'list', 'load'}
    elif curix == 2:
        if args[1] == 'load':
            possible = _list_installed_xontribs()
    else:
        raise StopIteration

    return {i for i in possible if i.startswith(prefix)}



# -*- coding: utf-8 -*-
"""Hooks for pygments syntax highlighting."""
import os
import re
import sys
import string
import builtins
from collections import ChainMap
from collections.abc import MutableMapping

# must come before pygments imports
from xonsh.lazyasd import load_module_in_background

from pygments.lexer import inherit, bygroups, include
from pygments.lexers.agile import PythonLexer
from pygments.token import (Keyword, Name, Comment, String, Error, Number,
                            Operator, Generic, Whitespace, Token, Punctuation,
                            Text)
from pygments.style import Style
from pygments.styles import get_style_by_name
import pygments.util

from xonsh.commands_cache import CommandsCache
from xonsh.lazyasd import LazyObject, LazyDict, lazyobject
from xonsh.tools import (ON_WINDOWS, intensify_colors_for_cmd_exe,
                         expand_gray_colors_for_cmd_exe)
from xonsh.color_tools import (RE_BACKGROUND, BASE_XONSH_COLORS, make_pallete,
                               find_closest_color)
from xonsh.style_tools import norm_name
from xonsh.lazyimps import terminal256
from xonsh.platform import os_environ

load_module_in_background('pkg_resources', debug='XONSH_DEBUG',
                          replacements={'pygments.plugin': 'pkg_resources'})


def _command_is_valid(cmd):
    try:
        cmd_abspath = os.path.abspath(os.path.expanduser(cmd))
    except (FileNotFoundError, OSError):
        return False
    return cmd in builtins.__xonsh_commands_cache__ or \
        (os.path.isfile(cmd_abspath) and os.access(cmd_abspath, os.X_OK))


def _command_is_autocd(cmd):
    if not builtins.__xonsh_env__.get('AUTO_CD', False):
        return False
    try:
        cmd_abspath = os.path.abspath(os.path.expanduser(cmd))
    except (FileNotFoundError, OSError):
        return False
    return os.path.isdir(cmd_abspath)


def subproc_cmd_callback(_, match):
    """Yield Builtin token if match contains valid command,
    otherwise fallback to fallback lexer.
    """
    cmd = match.group()
    yield match.start(), Name.Builtin if _command_is_valid(cmd) else Error, cmd


def subproc_arg_callback(_, match):
    """Check if match contains valid path"""
    text = match.group()
    try:
        ispath = os.path.exists(os.path.expanduser(text))
    except (FileNotFoundError, OSError):
        ispath = False
    yield (match.start(), Name.Constant if ispath else Text, text)


COMMAND_TOKEN_RE = r'[^=\s\[\]{}()$"\'`<&|;!]+(?=\s|$|\)|\]|\}|!)'


class XonshLexer(PythonLexer):
    """Xonsh console lexer for pygments."""

    name = 'Xonsh lexer'
    aliases = ['xonsh', 'xsh']
    filenames = ['*.xsh', '*xonshrc']

    def __init__(self, *args, **kwargs):
        # If the lexor is loaded as a pygment plugin, we have to mock
        # __xonsh_env__ and __xonsh_commands_cache__
        if not hasattr(builtins, '__xonsh_env__'):
            setattr(builtins, '__xonsh_env__', {})
            if ON_WINDOWS:
                pathext = os_environ.get('PATHEXT', ['.EXE', '.BAT', '.CMD'])
                builtins.__xonsh_env__['PATHEXT'] = pathext.split(os.pathsep)
        if not hasattr(builtins, '__xonsh_commands_cache__'):
            setattr(builtins, '__xonsh_commands_cache__', CommandsCache())
        _ = builtins.__xonsh_commands_cache__.all_commands  # NOQA
        super().__init__(*args, **kwargs)

    tokens = {
        'mode_switch_brackets': [
            (r'(\$)(\{)', bygroups(Keyword, Punctuation), 'py_curly_bracket'),
            (r'(@)(\()', bygroups(Keyword, Punctuation), 'py_bracket'),
            (r'([\!\$])(\()', bygroups(Keyword, Punctuation),
             ('subproc_bracket', 'subproc_start')),
            (r'(@\$)(\()', bygroups(Keyword, Punctuation),
             ('subproc_bracket', 'subproc_start')),
            (r'([\!\$])(\[)', bygroups(Keyword, Punctuation),
             ('subproc_square_bracket', 'subproc_start')),
            (r'(g?)(`)', bygroups(String.Affix, String.Backtick),
             'backtick_re'),
        ],
        'subproc_bracket': [
            (r'\)', Punctuation, '#pop'),
            include('subproc'),
        ],
        'subproc_square_bracket': [
            (r'\]', Punctuation, '#pop'),
            include('subproc'),
        ],
        'py_bracket': [
            (r'\)', Punctuation, '#pop'),
            include('root'),
        ],
        'py_curly_bracket': [
            (r'\}', Punctuation, '#pop'),
            include('root'),
        ],
        'backtick_re': [
            (r'[\.\^\$\*\+\?\[\]\|]', String.Regex),
            (r'({[0-9]+}|{[0-9]+,[0-9]+})\??', String.Regex),
            (r'\\([0-9]+|[AbBdDsSwWZabfnrtuUvx\\])', String.Escape),
            (r'`', String.Backtick, '#pop'),
            (r'[^`\.\^\$\*\+\?\[\]\|]+', String.Backtick),
        ],
        'root': [
            (r'\?', Keyword),
            (r'(?<=\w)!', Keyword),
            (r'\$\w+', Name.Variable),
            (r'\(', Punctuation, 'py_bracket'),
            (r'\{', Punctuation, 'py_curly_bracket'),
            include('mode_switch_brackets'),
            inherit,
        ],
        'subproc_start': [
            (r'\s+', Whitespace),
            (COMMAND_TOKEN_RE, subproc_cmd_callback, '#pop'),
            (r'', Whitespace, '#pop'),
        ],
        'subproc': [
            include('mode_switch_brackets'),
            (r'&&|\|\|', Operator, 'subproc_start'),
            (r'"(\\\\|\\[0-7]+|\\.|[^"\\])*"', String.Double),
            (r"'(\\\\|\\[0-7]+|\\.|[^'\\])*'", String.Single),
            (r'(?<=\w|\s)!', Keyword, 'subproc_macro'),
            (r'^!', Keyword, 'subproc_macro'),
            (r';', Punctuation, 'subproc_start'),
            (r'&|=', Punctuation),
            (r'\|', Punctuation, 'subproc_start'),
            (r'\s+', Text),
            (r'[^=\s\[\]{}()$"\'`<&|;]+', subproc_arg_callback),
            (r'<', Text),
            (r'\$\w+', Name.Variable),
        ],
        'subproc_macro': [
            (r'(\s*)([^\n]+)', bygroups(Whitespace, String)),
            (r'', Whitespace, '#pop'),
        ],
    }

    def get_tokens_unprocessed(self, text):
        """Check first command, then call super.get_tokens_unprocessed
        with root or subproc state"""
        start = 0
        state = ('root', )
        m = re.match(r'(\s*)({})'.format(COMMAND_TOKEN_RE), text)
        if m is not None:
            yield m.start(1), Whitespace, m.group(1)
            cmd = m.group(2)
            cmd_is_valid = _command_is_valid(cmd)
            cmd_is_autocd = _command_is_autocd(cmd)

            if cmd_is_valid or cmd_is_autocd:
                yield (m.start(2),
                       Name.Builtin if cmd_is_valid else Name.Constant,
                       cmd)
                start = m.end(2)
                state = ('subproc', )

        for i, t, v in super().get_tokens_unprocessed(text[start:], state):
            yield i + start, t, v


class XonshConsoleLexer(XonshLexer):
    """Xonsh console lexer for pygments."""

    name = 'Xonsh console lexer'
    aliases = ['xonshcon']
    filenames = []

    tokens = {
        'root': [(r'^(>>>|\.\.\.) ', Generic.Prompt),
                 (r'\n(>>>|\.\.\.)', Generic.Prompt),
                 (r'\n(?![>.][>.][>.] )([^\n]*)', Generic.Output),
                 (r'\n(?![>.][>.][>.] )(.*?)$', Generic.Output),
                 inherit]
    }


#
# Colors and Styles
#

Color = Token.Color  # alias to new color token namespace


def color_by_name(name, fg=None, bg=None):
    """Converts a color name to a color token, foreground name,
    and background name.  Will take into consideration current foreground
    and background colors, if provided.

    Parameters
    ----------
    name : str
        Color name.
    fg : str, optional
        Foreground color name.
    bg : str, optional
        Background color name.

    Returns
    -------
    tok : Token
        Pygments Token.Color subclass
    fg : str or None
        New computed foreground color name.
    bg : str or None
        New computed background color name.
    """
    name = name.upper()
    if name == 'NO_COLOR':
        return Color.NO_COLOR, None, None
    m = RE_BACKGROUND.search(name)
    if m is None:  # must be foreground color
        fg = norm_name(name)
    else:
        bg = norm_name(name)
    # assmble token
    if fg is None and bg is None:
        tokname = 'NO_COLOR'
    elif fg is None:
        tokname = bg
    elif bg is None:
        tokname = fg
    else:
        tokname = fg + '__' + bg
    tok = getattr(Color, tokname)
    return tok, fg, bg


def code_by_name(name, styles):
    """Converts a token name into a pygments-style color code.

    Parameters
    ----------
    name : str
        Color token name.
    styles : Mapping
        Mapping for looking up non-hex colors

    Returns
    -------
    code : str
        Pygments style color code.
    """
    fg, _, bg = name.lower().partition('__')
    if fg.startswith('background_'):
        fg, bg = bg, fg
    codes = []
    # foreground color
    if len(fg) == 0:
        pass
    elif 'hex' in fg:
        for p in fg.split('_'):
            codes.append('#'+p[3:] if p.startswith('hex') else p)
    else:
        fgtok = getattr(Color, fg.upper())
        if fgtok in styles:
            codes.append(styles[fgtok])
        else:
            codes += fg.split('_')
    # background color
    if len(bg) == 0:
        pass
    elif bg.startswith('background_hex'):
        codes.append('bg:#'+bg[14:])
    else:
        bgtok = getattr(Color, bg.upper())
        if bgtok in styles:
            codes.append(styles[bgtok])
        else:
            codes.append(bg.replace('background_', 'bg:'))
    code = ' '.join(codes)
    return code


def partial_color_tokenize(template):
    """Tokenizes a template string containing colors. Will return a list
    of tuples mapping the token to the string which has that color.
    These sub-strings maybe templates themselves.
    """
    if hasattr(builtins, '__xonsh_shell__'):
        styles = __xonsh_shell__.shell.styler.styles
    else:
        styles = None
    color = Color.NO_COLOR
    try:
        toks, color = _partial_color_tokenize_main(template, styles)
    except Exception:
        toks = [(Color.NO_COLOR, template)]
    if styles is not None:
        styles[color]  # ensure color is available
    return toks


def _partial_color_tokenize_main(template, styles):
    formatter = string.Formatter()
    bopen = '{'
    bclose = '}'
    colon = ':'
    expl = '!'
    color = Color.NO_COLOR
    fg = bg = None
    value = ''
    toks = []
    for literal, field, spec, conv in formatter.parse(template):
        if field is None:
            value += literal
        elif field in KNOWN_COLORS or '#' in field:
            value += literal
            next_color, fg, bg = color_by_name(field, fg, bg)
            if next_color is not color:
                if len(value) > 0:
                    toks.append((color, value))
                    if styles is not None:
                        styles[color]  # ensure color is available
                color = next_color
                value = ''
        elif field is not None:
            parts = [literal, bopen, field]
            if conv is not None and len(conv) > 0:
                parts.append(expl)
                parts.append(conv)
            if spec is not None and len(spec) > 0:
                parts.append(colon)
                parts.append(spec)
            parts.append(bclose)
            value += ''.join(parts)
        else:
            value += literal
    toks.append((color, value))
    return toks, color


class CompoundColorMap(MutableMapping):
    """Looks up color tokes by name, potentailly generating the value
    from the lookup.
    """

    def __init__(self, styles, *args, **kwargs):
        self.styles = styles
        self.colors = dict(*args, **kwargs)

    def __getitem__(self, key):
        if key in self.colors:
            return self.colors[key]
        if key in self.styles:
            value = self.styles[key]
            self[key] = value
            return value
        if key is Color:
            raise KeyError
        pre, _, name = str(key).rpartition('.')
        if pre != 'Token.Color':
            raise KeyError
        value = code_by_name(name, self.styles)
        self[key] = value
        return value

    def __setitem__(self, key, value):
        self.colors[key] = value

    def __delitem__(self, key):
        del self.colors[key]

    def __iter__(self):
        yield from self.colors.keys()

    def __len__(self):
        return len(self.colors)


class XonshStyle(Style):
    """A xonsh pygments style that will dispatch to the correct color map
    by using a ChainMap.  The style_name property may be used to reset
    the current style.
    """

    def __init__(self, style_name='default'):
        """
        Parameters
        ----------
        style_name : str, optional
            The style name to initialize with.
        """
        self.trap = {}  # for traping custom colors set by user
        self._smap = {}
        self._style_name = ''
        self.style_name = style_name
        super().__init__()

    @property
    def style_name(self):
        return self._style_name

    @style_name.setter
    def style_name(self, value):
        if self._style_name == value:
            return
        if value not in STYLES:
            try:  # loading style dynamically
                pygments_style_by_name(value)
            except Exception:
                print('Could not find style {0!r}, using default'.format(value),
                      file=sys.stderr)
                value = 'default'
                builtins.__xonsh_env__['XONSH_COLOR_STYLE'] = value
        cmap = STYLES[value]
        if value == 'default':
            self._smap = XONSH_BASE_STYLE.copy()
        else:
            try:
                self._smap = get_style_by_name(value)().styles.copy()
            except (ImportError, pygments.util.ClassNotFound):
                self._smap = XONSH_BASE_STYLE.copy()
        compound = CompoundColorMap(ChainMap(self.trap, cmap, PTK_STYLE, self._smap))
        self.styles = ChainMap(self.trap, cmap, PTK_STYLE, self._smap, compound)
        self._style_name = value
        if ON_WINDOWS:
            self.enhance_colors_for_cmd_exe()

    @style_name.deleter
    def style_name(self):
        self._style_name = ''

    def enhance_colors_for_cmd_exe(self):
        """ Enhance colors when using cmd.exe on windows.
            When using the default style all blue and dark red colors
            are changed to CYAN and intence red.
        """
        env = builtins.__xonsh_env__
        # Ensure we are not using ConEmu
        if 'CONEMUANSI' not in env:
            if not hasattr(pygments.style, 'ansicolors'):
                # Auto suggest needs to be a darker shade to be distinguishable
                # from the default color
                self.styles[Token.AutoSuggestion] = '#444444'
                self._smap.update(expand_gray_colors_for_cmd_exe(self._smap))
            if env.get('INTENSIFY_COLORS_ON_WIN', False):
                has_ansi = hasattr(pygments.style, 'ansicolors')
                newcolors = intensify_colors_for_cmd_exe(self._smap, ansi=has_ansi)
                self._smap.update(newcolors)


def xonsh_style_proxy(styler):
    """Factory for a proxy class to a xonsh style."""
    class XonshStyleProxy(Style):
        """Simple proxy class to fool prompt toolkit."""

        target = styler
        styles = styler.styles

        def __new__(cls, *args, **kwargs):
            return cls.target

    return XonshStyleProxy


if hasattr(pygments.style, 'ansicolors'):
    PTK_STYLE = {
        Token.Menu.Completions: 'bg:#ansilightgray #ansiblack',
        Token.Menu.Completions.Completion: '',
        Token.Menu.Completions.Completion.Current: 'bg:#ansidarkgray #ansiwhite',
        Token.Scrollbar: 'bg:#ansidarkgray',
        Token.Scrollbar.Button: 'bg:#ansiblack',
        Token.Scrollbar.Arrow: 'bg:#ansiblack #ansiwhite bold',
        Token.AutoSuggestion: '#ansidarkgray',
        Token.Aborted: '#ansidarkgray',
    }
else:
    PTK_STYLE = {
        Token.Menu.Completions: 'bg:#888888 #000000',
        Token.Menu.Completions.Completion: '',
        Token.Menu.Completions.Completion.Current: 'bg:#555555 #ffffff',
        Token.Scrollbar: 'bg:#555555',
        Token.Scrollbar.Button: 'bg:#000000',
        Token.Scrollbar.Arrow: 'bg:#000000 #ffffff bold',
        Token.AutoSuggestion: '#666666',
        Token.Aborted: '#666666',
    }

if hasattr(pygments.style, 'ansicolors'):
    XONSH_BASE_STYLE = LazyObject(lambda: {
        Whitespace: '#ansilightgray',
        Comment: 'underline #ansiteal',
        Comment.Preproc: 'underline #ansibrown',
        Keyword: 'bold #ansidarkgreen',
        Keyword.Pseudo: 'nobold',
        Keyword.Type: 'nobold #ansidarkred',
        Operator: '#ansidarkgray',
        Operator.Word: 'bold #ansipurple',
        Name.Builtin: '#ansidarkgreen',
        Name.Function: '#ansiblue',
        Name.Class: 'bold #ansiblue',
        Name.Namespace: 'bold #ansiblue',
        Name.Exception: 'bold #ansired',
        Name.Variable: '#ansidarkblue',
        Name.Constant: '#ansidarkred',
        Name.Label: '#ansiyellow',
        Name.Entity: 'bold #ansilightgray',
        Name.Attribute: '#ansiyellow',
        Name.Tag: 'bold #ansidarkgreen',
        Name.Decorator: '#ansifuchsia',
        String: '#ansired',
        String.Doc: 'underline',
        String.Interpol: 'bold #ansipurple',
        String.Escape: 'bold #ansibrown',
        String.Regex: '#ansipurple',
        String.Symbol: '#ansibrown',
        String.Other: '#ansidarkgreen',
        Number: '#ansidarkgray',
        Generic.Heading: 'bold #ansidarkblue',
        Generic.Subheading: 'bold #ansipurple',
        Generic.Deleted: '#ansidarkred',
        Generic.Inserted: '#ansigreen',
        Generic.Error: 'bold #ansired',
        Generic.Emph: 'underline',
        Generic.Prompt: 'bold #ansidarkblue',
        Generic.Output: '#ansidarkblue',
        Generic.Traceback: '#ansidarkblue',
        Error: '#ansired',
        }, globals(), 'XONSH_BASE_STYLE')
else:
    XONSH_BASE_STYLE = LazyObject(lambda: {
        Whitespace: "#bbbbbb",
        Comment: "italic #408080",
        Comment.Preproc: "noitalic #BC7A00",
        Keyword: "bold #008000",
        Keyword.Pseudo: "nobold",
        Keyword.Type:   "nobold #B00040",
        Operator:   "#666666",
        Operator.Word:  "bold #AA22FF",
        Name.Builtin: "#008000",
        Name.Function: "#0000FF",
        Name.Class: "bold #0000FF",
        Name.Namespace: "bold #0000FF",
        Name.Exception: "bold #D2413A",
        Name.Variable: "#19177C",
        Name.Constant: "#880000",
        Name.Label: "#A0A000",
        Name.Entity: "bold #999999",
        Name.Attribute: "#7D9029",
        Name.Tag: "bold #008000",
        Name.Decorator: "#AA22FF",
        String: "#BA2121",
        String.Doc: "italic",
        String.Interpol: "bold #BB6688",
        String.Escape: "bold #BB6622",
        String.Regex: "#BB6688",
        String.Symbol: "#19177C",
        String.Other: "#008000",
        Number: "#666666",
        Generic.Heading: "bold #000080",
        Generic.Subheading: "bold #800080",
        Generic.Deleted: "#A00000",
        Generic.Inserted: "#00A000",
        Generic.Error: "#FF0000",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.Prompt: "bold #000080",
        Generic.Output: "#888",
        Generic.Traceback: "#04D",
        Error: "border:#FF0000"
    }, globals(), 'XONSH_BASE_STYLE')


KNOWN_COLORS = LazyObject(lambda: frozenset([
    'BACKGROUND_BLACK',
    'BACKGROUND_BLUE',
    'BACKGROUND_CYAN',
    'BACKGROUND_GREEN',
    'BACKGROUND_INTENSE_BLACK',
    'BACKGROUND_INTENSE_BLUE',
    'BACKGROUND_INTENSE_CYAN',
    'BACKGROUND_INTENSE_GREEN',
    'BACKGROUND_INTENSE_PURPLE',
    'BACKGROUND_INTENSE_RED',
    'BACKGROUND_INTENSE_WHITE',
    'BACKGROUND_INTENSE_YELLOW',
    'BACKGROUND_PURPLE',
    'BACKGROUND_RED',
    'BACKGROUND_WHITE',
    'BACKGROUND_YELLOW',
    'BLACK',
    'BLUE',
    'BOLD_BLACK',
    'BOLD_BLUE',
    'BOLD_CYAN',
    'BOLD_GREEN',
    'BOLD_INTENSE_BLACK',
    'BOLD_INTENSE_BLUE',
    'BOLD_INTENSE_CYAN',
    'BOLD_INTENSE_GREEN',
    'BOLD_INTENSE_PURPLE',
    'BOLD_INTENSE_RED',
    'BOLD_INTENSE_WHITE',
    'BOLD_INTENSE_YELLOW',
    'BOLD_PURPLE',
    'BOLD_RED',
    'BOLD_UNDERLINE_BLACK',
    'BOLD_UNDERLINE_BLUE',
    'BOLD_UNDERLINE_CYAN',
    'BOLD_UNDERLINE_GREEN',
    'BOLD_UNDERLINE_INTENSE_BLACK',
    'BOLD_UNDERLINE_INTENSE_BLUE',
    'BOLD_UNDERLINE_INTENSE_CYAN',
    'BOLD_UNDERLINE_INTENSE_GREEN',
    'BOLD_UNDERLINE_INTENSE_PURPLE',
    'BOLD_UNDERLINE_INTENSE_RED',
    'BOLD_UNDERLINE_INTENSE_WHITE',
    'BOLD_UNDERLINE_INTENSE_YELLOW',
    'BOLD_UNDERLINE_PURPLE',
    'BOLD_UNDERLINE_RED',
    'BOLD_UNDERLINE_WHITE',
    'BOLD_UNDERLINE_YELLOW',
    'BOLD_WHITE',
    'BOLD_YELLOW',
    'CYAN',
    'GREEN',
    'INTENSE_BLACK',
    'INTENSE_BLUE',
    'INTENSE_CYAN',
    'INTENSE_GREEN',
    'INTENSE_PURPLE',
    'INTENSE_RED',
    'INTENSE_WHITE',
    'INTENSE_YELLOW',
    'NO_COLOR',
    'PURPLE',
    'RED',
    'UNDERLINE_BLACK',
    'UNDERLINE_BLUE',
    'UNDERLINE_CYAN',
    'UNDERLINE_GREEN',
    'UNDERLINE_INTENSE_BLACK',
    'UNDERLINE_INTENSE_BLUE',
    'UNDERLINE_INTENSE_CYAN',
    'UNDERLINE_INTENSE_GREEN',
    'UNDERLINE_INTENSE_PURPLE',
    'UNDERLINE_INTENSE_RED',
    'UNDERLINE_INTENSE_WHITE',
    'UNDERLINE_INTENSE_YELLOW',
    'UNDERLINE_PURPLE',
    'UNDERLINE_RED',
    'UNDERLINE_WHITE',
    'UNDERLINE_YELLOW',
    'WHITE',
    'YELLOW',
    ]), globals(), 'KNOWN_COLORS')


def _expand_style(cmap):
    """Expands a style in order to more quickly make color map changes."""
    for key, val in list(cmap.items()):
        if key is Color.NO_COLOR:
            continue
        _, _, key = str(key).rpartition('.')
        cmap[getattr(Color, 'BOLD_'+key)] = 'bold ' + val
        cmap[getattr(Color, 'UNDERLINE_'+key)] = 'underline ' + val
        cmap[getattr(Color, 'BOLD_UNDERLINE_'+key)] = 'bold underline ' + val
        if val == 'noinherit':
            cmap[getattr(Color, 'BACKGROUND_'+key)] = val
        else:
            cmap[getattr(Color, 'BACKGROUND_'+key)] = 'bg:' + val


def _bw_style():
    style = {
        Color.BLACK: 'noinherit',
        Color.BLUE: 'noinherit',
        Color.CYAN: 'noinherit',
        Color.GREEN: 'noinherit',
        Color.INTENSE_BLACK: 'noinherit',
        Color.INTENSE_BLUE: 'noinherit',
        Color.INTENSE_CYAN: 'noinherit',
        Color.INTENSE_GREEN: 'noinherit',
        Color.INTENSE_PURPLE: 'noinherit',
        Color.INTENSE_RED: 'noinherit',
        Color.INTENSE_WHITE: 'noinherit',
        Color.INTENSE_YELLOW: 'noinherit',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: 'noinherit',
        Color.RED: 'noinherit',
        Color.WHITE: 'noinherit',
        Color.YELLOW: 'noinherit',
        }
    _expand_style(style)
    return style


def _default_style():
    if hasattr(pygments.style, 'ansicolors'):
        style = {
            Color.BLACK: '#ansiblack',
            Color.BLUE: '#ansidarkblue',
            Color.CYAN: '#ansiteal',
            Color.GREEN: '#ansidarkgreen',
            Color.INTENSE_BLACK: '#ansidarkgray',
            Color.INTENSE_BLUE: '#ansiblue',
            Color.INTENSE_CYAN: '#ansiturquoise',
            Color.INTENSE_GREEN: '#ansigreen',
            Color.INTENSE_PURPLE: '#ansifuchsia',
            Color.INTENSE_RED: '#ansired',
            Color.INTENSE_WHITE: '#ansiwhite',
            Color.INTENSE_YELLOW: '#ansiyellow',
            Color.NO_COLOR: 'noinherit',
            Color.PURPLE: '#ansipurple',
            Color.RED: '#ansidarkred',
            Color.WHITE: '#ansilightgray',
            Color.YELLOW: '#ansibrown',
        }
    elif ON_WINDOWS and 'CONEMUANSI' not in os_environ:
        # These colors must match the color specification
        # in prompt_toolkit, so the colors are converted
        # correctly when using cmd.exe
        style = {
            Color.BLACK: '#000000',
            Color.BLUE: '#0000AA',
            Color.CYAN: '#00AAAA',
            Color.GREEN: '#00AA00',
            Color.INTENSE_BLACK: '#444444',
            Color.INTENSE_BLUE: '#4444FF',
            Color.INTENSE_CYAN: '#44FFFF',
            Color.INTENSE_GREEN: '#44FF44',
            Color.INTENSE_PURPLE: '#FF44FF',
            Color.INTENSE_RED: '#FF4444',
            Color.INTENSE_WHITE: '#FFFFFF',
            Color.INTENSE_YELLOW: '#FFFF44',
            Color.NO_COLOR: 'noinherit',
            Color.PURPLE: '#AA00AA',
            Color.RED: '#AA0000',
            Color.WHITE: '#888888',
            Color.YELLOW: '#AAAA00',
        }
    else:
        style = {
            Color.BLACK: '#000000',
            Color.BLUE: '#0000AA',
            Color.CYAN: '#00AAAA',
            Color.GREEN: '#00AA00',
            Color.INTENSE_BLACK: '#555555',
            Color.INTENSE_BLUE: '#0000FF',
            Color.INTENSE_CYAN: '#55FFFF',
            Color.INTENSE_GREEN: '#00FF00',
            Color.INTENSE_PURPLE: '#FF00FF',
            Color.INTENSE_RED: '#FF0000',
            Color.INTENSE_WHITE: '#ffffff',
            Color.INTENSE_YELLOW: '#FFFF55',
            Color.NO_COLOR: 'noinherit',
            Color.PURPLE: '#AA00AA',
            Color.RED: '#AA0000',
            Color.WHITE: '#aaaaaa',
            Color.YELLOW: '#ffff00',
        }
    _expand_style(style)
    return style


def _monokai_style():
    style = {
        Color.BLACK: '#1e0010',
        Color.BLUE: '#6666ef',
        Color.CYAN: '#66d9ef',
        Color.GREEN: '#2ee22e',
        Color.INTENSE_BLACK: '#5e5e5e',
        Color.INTENSE_BLUE: '#2626d7',
        Color.INTENSE_CYAN: '#2ed9d9',
        Color.INTENSE_GREEN: '#a6e22e',
        Color.INTENSE_PURPLE: '#ae81ff',
        Color.INTENSE_RED: '#f92672',
        Color.INTENSE_WHITE: '#f8f8f2',
        Color.INTENSE_YELLOW: '#e6db74',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#960050',
        Color.RED: '#AF0000',
        Color.WHITE: '#d7d7d7',
        Color.YELLOW: '#e2e22e',
    }
    _expand_style(style)
    return style


######################################
#   Auto-generated below this line   #
######################################
def _algol_style():
    style = {
        Color.BLACK: '#666',
        Color.BLUE: '#666',
        Color.CYAN: '#666',
        Color.GREEN: '#666',
        Color.INTENSE_BLACK: '#666',
        Color.INTENSE_BLUE: '#888',
        Color.INTENSE_CYAN: '#888',
        Color.INTENSE_GREEN: '#888',
        Color.INTENSE_PURPLE: '#888',
        Color.INTENSE_RED: '#FF0000',
        Color.INTENSE_WHITE: '#888',
        Color.INTENSE_YELLOW: '#888',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#666',
        Color.RED: '#FF0000',
        Color.WHITE: '#888',
        Color.YELLOW: '#FF0000',
        }
    _expand_style(style)
    return style


def _algol_nu_style():
    style = {
        Color.BLACK: '#666',
        Color.BLUE: '#666',
        Color.CYAN: '#666',
        Color.GREEN: '#666',
        Color.INTENSE_BLACK: '#666',
        Color.INTENSE_BLUE: '#888',
        Color.INTENSE_CYAN: '#888',
        Color.INTENSE_GREEN: '#888',
        Color.INTENSE_PURPLE: '#888',
        Color.INTENSE_RED: '#FF0000',
        Color.INTENSE_WHITE: '#888',
        Color.INTENSE_YELLOW: '#888',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#666',
        Color.RED: '#FF0000',
        Color.WHITE: '#888',
        Color.YELLOW: '#FF0000',
        }
    _expand_style(style)
    return style


def _autumn_style():
    style = {
        Color.BLACK: '#000080',
        Color.BLUE: '#0000aa',
        Color.CYAN: '#00aaaa',
        Color.GREEN: '#00aa00',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#1e90ff',
        Color.INTENSE_CYAN: '#1e90ff',
        Color.INTENSE_GREEN: '#4c8317',
        Color.INTENSE_PURPLE: '#FAA',
        Color.INTENSE_RED: '#aa5500',
        Color.INTENSE_WHITE: '#bbbbbb',
        Color.INTENSE_YELLOW: '#FAA',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#aa0000',
        Color.WHITE: '#aaaaaa',
        Color.YELLOW: '#aa5500',
        }
    _expand_style(style)
    return style


def _borland_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#000080',
        Color.CYAN: '#008080',
        Color.GREEN: '#008800',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#0000FF',
        Color.INTENSE_CYAN: '#ddffdd',
        Color.INTENSE_GREEN: '#888888',
        Color.INTENSE_PURPLE: '#e3d2d2',
        Color.INTENSE_RED: '#FF0000',
        Color.INTENSE_WHITE: '#ffdddd',
        Color.INTENSE_YELLOW: '#e3d2d2',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#aa0000',
        Color.WHITE: '#aaaaaa',
        Color.YELLOW: '#a61717',
        }
    _expand_style(style)
    return style


def _colorful_style():
    style = {
        Color.BLACK: '#000',
        Color.BLUE: '#00C',
        Color.CYAN: '#0e84b5',
        Color.GREEN: '#00A000',
        Color.INTENSE_BLACK: '#555',
        Color.INTENSE_BLUE: '#33B',
        Color.INTENSE_CYAN: '#bbbbbb',
        Color.INTENSE_GREEN: '#888',
        Color.INTENSE_PURPLE: '#FAA',
        Color.INTENSE_RED: '#D42',
        Color.INTENSE_WHITE: '#fff0ff',
        Color.INTENSE_YELLOW: '#FAA',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#A00000',
        Color.WHITE: '#bbbbbb',
        Color.YELLOW: '#A60',
        }
    _expand_style(style)
    return style


def _emacs_style():
    style = {
        Color.BLACK: '#008000',
        Color.BLUE: '#000080',
        Color.CYAN: '#04D',
        Color.GREEN: '#00A000',
        Color.INTENSE_BLACK: '#666666',
        Color.INTENSE_BLUE: '#04D',
        Color.INTENSE_CYAN: '#bbbbbb',
        Color.INTENSE_GREEN: '#00BB00',
        Color.INTENSE_PURPLE: '#AA22FF',
        Color.INTENSE_RED: '#D2413A',
        Color.INTENSE_WHITE: '#bbbbbb',
        Color.INTENSE_YELLOW: '#bbbbbb',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#A00000',
        Color.WHITE: '#bbbbbb',
        Color.YELLOW: '#BB6622',
        }
    _expand_style(style)
    return style


def _friendly_style():
    style = {
        Color.BLACK: '#007020',
        Color.BLUE: '#000080',
        Color.CYAN: '#0e84b5',
        Color.GREEN: '#00A000',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#70a0d0',
        Color.INTENSE_CYAN: '#60add5',
        Color.INTENSE_GREEN: '#40a070',
        Color.INTENSE_PURPLE: '#bb60d5',
        Color.INTENSE_RED: '#d55537',
        Color.INTENSE_WHITE: '#fff0f0',
        Color.INTENSE_YELLOW: '#bbbbbb',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#A00000',
        Color.WHITE: '#bbbbbb',
        Color.YELLOW: '#c65d09',
        }
    _expand_style(style)
    return style


def _fruity_style():
    style = {
        Color.BLACK: '#0f140f',
        Color.BLUE: '#0086d2',
        Color.CYAN: '#0086d2',
        Color.GREEN: '#008800',
        Color.INTENSE_BLACK: '#444444',
        Color.INTENSE_BLUE: '#0086f7',
        Color.INTENSE_CYAN: '#0086f7',
        Color.INTENSE_GREEN: '#888888',
        Color.INTENSE_PURPLE: '#ff0086',
        Color.INTENSE_RED: '#fb660a',
        Color.INTENSE_WHITE: '#ffffff',
        Color.INTENSE_YELLOW: '#cdcaa9',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#ff0086',
        Color.RED: '#ff0007',
        Color.WHITE: '#cdcaa9',
        Color.YELLOW: '#fb660a',
        }
    _expand_style(style)
    return style


def _igor_style():
    style = {
        Color.BLACK: '#009C00',
        Color.BLUE: '#0000FF',
        Color.CYAN: '#007575',
        Color.GREEN: '#009C00',
        Color.INTENSE_BLACK: '#007575',
        Color.INTENSE_BLUE: '#0000FF',
        Color.INTENSE_CYAN: '#007575',
        Color.INTENSE_GREEN: '#009C00',
        Color.INTENSE_PURPLE: '#CC00A3',
        Color.INTENSE_RED: '#C34E00',
        Color.INTENSE_WHITE: '#CC00A3',
        Color.INTENSE_YELLOW: '#C34E00',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#CC00A3',
        Color.RED: '#C34E00',
        Color.WHITE: '#CC00A3',
        Color.YELLOW: '#C34E00',
        }
    _expand_style(style)
    return style


def _lovelace_style():
    style = {
        Color.BLACK: '#444444',
        Color.BLUE: '#2838b0',
        Color.CYAN: '#289870',
        Color.GREEN: '#388038',
        Color.INTENSE_BLACK: '#666666',
        Color.INTENSE_BLUE: '#2838b0',
        Color.INTENSE_CYAN: '#888888',
        Color.INTENSE_GREEN: '#289870',
        Color.INTENSE_PURPLE: '#a848a8',
        Color.INTENSE_RED: '#b83838',
        Color.INTENSE_WHITE: '#888888',
        Color.INTENSE_YELLOW: '#a89028',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#a848a8',
        Color.RED: '#c02828',
        Color.WHITE: '#888888',
        Color.YELLOW: '#b85820',
        }
    _expand_style(style)
    return style


def _manni_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#000099',
        Color.CYAN: '#009999',
        Color.GREEN: '#00CC00',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#9999FF',
        Color.INTENSE_CYAN: '#00CCFF',
        Color.INTENSE_GREEN: '#99CC66',
        Color.INTENSE_PURPLE: '#CC00FF',
        Color.INTENSE_RED: '#FF6600',
        Color.INTENSE_WHITE: '#FFCCCC',
        Color.INTENSE_YELLOW: '#FFCC33',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#CC00FF',
        Color.RED: '#AA0000',
        Color.WHITE: '#AAAAAA',
        Color.YELLOW: '#CC3300',
        }
    _expand_style(style)
    return style


def _murphy_style():
    style = {
        Color.BLACK: '#000',
        Color.BLUE: '#000080',
        Color.CYAN: '#0e84b5',
        Color.GREEN: '#00A000',
        Color.INTENSE_BLACK: '#555',
        Color.INTENSE_BLUE: '#66f',
        Color.INTENSE_CYAN: '#5ed',
        Color.INTENSE_GREEN: '#5ed',
        Color.INTENSE_PURPLE: '#e9e',
        Color.INTENSE_RED: '#f84',
        Color.INTENSE_WHITE: '#eee',
        Color.INTENSE_YELLOW: '#fc8',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#A00000',
        Color.WHITE: '#bbbbbb',
        Color.YELLOW: '#c65d09',
        }
    _expand_style(style)
    return style


def _native_style():
    style = {
        Color.BLACK: '#520000',
        Color.BLUE: '#3677a9',
        Color.CYAN: '#24909d',
        Color.GREEN: '#589819',
        Color.INTENSE_BLACK: '#666666',
        Color.INTENSE_BLUE: '#447fcf',
        Color.INTENSE_CYAN: '#40ffff',
        Color.INTENSE_GREEN: '#6ab825',
        Color.INTENSE_PURPLE: '#e3d2d2',
        Color.INTENSE_RED: '#cd2828',
        Color.INTENSE_WHITE: '#ffffff',
        Color.INTENSE_YELLOW: '#ed9d13',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#666666',
        Color.RED: '#a61717',
        Color.WHITE: '#aaaaaa',
        Color.YELLOW: '#a61717',
        }
    _expand_style(style)
    return style


def _paraiso_dark_style():
    style = {
        Color.BLACK: '#776e71',
        Color.BLUE: '#815ba4',
        Color.CYAN: '#06b6ef',
        Color.GREEN: '#48b685',
        Color.INTENSE_BLACK: '#776e71',
        Color.INTENSE_BLUE: '#815ba4',
        Color.INTENSE_CYAN: '#5bc4bf',
        Color.INTENSE_GREEN: '#48b685',
        Color.INTENSE_PURPLE: '#e7e9db',
        Color.INTENSE_RED: '#ef6155',
        Color.INTENSE_WHITE: '#e7e9db',
        Color.INTENSE_YELLOW: '#fec418',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#815ba4',
        Color.RED: '#ef6155',
        Color.WHITE: '#5bc4bf',
        Color.YELLOW: '#f99b15',
        }
    _expand_style(style)
    return style


def _paraiso_light_style():
    style = {
        Color.BLACK: '#2f1e2e',
        Color.BLUE: '#2f1e2e',
        Color.CYAN: '#06b6ef',
        Color.GREEN: '#48b685',
        Color.INTENSE_BLACK: '#2f1e2e',
        Color.INTENSE_BLUE: '#815ba4',
        Color.INTENSE_CYAN: '#5bc4bf',
        Color.INTENSE_GREEN: '#48b685',
        Color.INTENSE_PURPLE: '#815ba4',
        Color.INTENSE_RED: '#ef6155',
        Color.INTENSE_WHITE: '#5bc4bf',
        Color.INTENSE_YELLOW: '#fec418',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#815ba4',
        Color.RED: '#2f1e2e',
        Color.WHITE: '#8d8687',
        Color.YELLOW: '#f99b15',
        }
    _expand_style(style)
    return style


def _pastie_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#0000DD',
        Color.CYAN: '#0066bb',
        Color.GREEN: '#008800',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#3333bb',
        Color.INTENSE_CYAN: '#ddffdd',
        Color.INTENSE_GREEN: '#22bb22',
        Color.INTENSE_PURPLE: '#e3d2d2',
        Color.INTENSE_RED: '#dd7700',
        Color.INTENSE_WHITE: '#fff0ff',
        Color.INTENSE_YELLOW: '#e3d2d2',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#bb0066',
        Color.RED: '#aa0000',
        Color.WHITE: '#bbbbbb',
        Color.YELLOW: '#aa6600',
        }
    _expand_style(style)
    return style


def _perldoc_style():
    style = {
        Color.BLACK: '#000080',
        Color.BLUE: '#000080',
        Color.CYAN: '#1e889b',
        Color.GREEN: '#00aa00',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#B452CD',
        Color.INTENSE_CYAN: '#bbbbbb',
        Color.INTENSE_GREEN: '#228B22',
        Color.INTENSE_PURPLE: '#B452CD',
        Color.INTENSE_RED: '#CD5555',
        Color.INTENSE_WHITE: '#e3d2d2',
        Color.INTENSE_YELLOW: '#e3d2d2',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#8B008B',
        Color.RED: '#aa0000',
        Color.WHITE: '#a7a7a7',
        Color.YELLOW: '#cb6c20',
        }
    _expand_style(style)
    return style


def _rrt_style():
    style = {
        Color.BLACK: '#ff0000',
        Color.BLUE: '#87ceeb',
        Color.CYAN: '#87ceeb',
        Color.GREEN: '#00ff00',
        Color.INTENSE_BLACK: '#87ceeb',
        Color.INTENSE_BLUE: '#87ceeb',
        Color.INTENSE_CYAN: '#7fffd4',
        Color.INTENSE_GREEN: '#00ff00',
        Color.INTENSE_PURPLE: '#ee82ee',
        Color.INTENSE_RED: '#ff0000',
        Color.INTENSE_WHITE: '#e5e5e5',
        Color.INTENSE_YELLOW: '#eedd82',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#ee82ee',
        Color.RED: '#ff0000',
        Color.WHITE: '#87ceeb',
        Color.YELLOW: '#ff0000',
        }
    _expand_style(style)
    return style


def _tango_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#0000cf',
        Color.CYAN: '#3465a4',
        Color.GREEN: '#00A000',
        Color.INTENSE_BLACK: '#204a87',
        Color.INTENSE_BLUE: '#5c35cc',
        Color.INTENSE_CYAN: '#f8f8f8',
        Color.INTENSE_GREEN: '#4e9a06',
        Color.INTENSE_PURPLE: '#f8f8f8',
        Color.INTENSE_RED: '#ef2929',
        Color.INTENSE_WHITE: '#f8f8f8',
        Color.INTENSE_YELLOW: '#c4a000',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#a40000',
        Color.WHITE: '#f8f8f8',
        Color.YELLOW: '#8f5902',
        }
    _expand_style(style)
    return style


def _trac_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#000080',
        Color.CYAN: '#009999',
        Color.GREEN: '#808000',
        Color.INTENSE_BLACK: '#555555',
        Color.INTENSE_BLUE: '#445588',
        Color.INTENSE_CYAN: '#ddffdd',
        Color.INTENSE_GREEN: '#999988',
        Color.INTENSE_PURPLE: '#e3d2d2',
        Color.INTENSE_RED: '#bb8844',
        Color.INTENSE_WHITE: '#ffdddd',
        Color.INTENSE_YELLOW: '#e3d2d2',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#800080',
        Color.RED: '#aa0000',
        Color.WHITE: '#aaaaaa',
        Color.YELLOW: '#808000',
        }
    _expand_style(style)
    return style


def _vim_style():
    style = {
        Color.BLACK: '#000080',
        Color.BLUE: '#000080',
        Color.CYAN: '#00cdcd',
        Color.GREEN: '#00cd00',
        Color.INTENSE_BLACK: '#666699',
        Color.INTENSE_BLUE: '#3399cc',
        Color.INTENSE_CYAN: '#00cdcd',
        Color.INTENSE_GREEN: '#00cd00',
        Color.INTENSE_PURPLE: '#cd00cd',
        Color.INTENSE_RED: '#FF0000',
        Color.INTENSE_WHITE: '#cccccc',
        Color.INTENSE_YELLOW: '#cdcd00',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#cd00cd',
        Color.RED: '#cd0000',
        Color.WHITE: '#cccccc',
        Color.YELLOW: '#cd0000',
        }
    _expand_style(style)
    return style


def _vs_style():
    style = {
        Color.BLACK: '#008000',
        Color.BLUE: '#0000ff',
        Color.CYAN: '#2b91af',
        Color.GREEN: '#008000',
        Color.INTENSE_BLACK: '#2b91af',
        Color.INTENSE_BLUE: '#2b91af',
        Color.INTENSE_CYAN: '#2b91af',
        Color.INTENSE_GREEN: '#2b91af',
        Color.INTENSE_PURPLE: '#2b91af',
        Color.INTENSE_RED: '#FF0000',
        Color.INTENSE_WHITE: '#2b91af',
        Color.INTENSE_YELLOW: '#2b91af',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#a31515',
        Color.RED: '#a31515',
        Color.WHITE: '#2b91af',
        Color.YELLOW: '#a31515',
        }
    _expand_style(style)
    return style


def _xcode_style():
    style = {
        Color.BLACK: '#000000',
        Color.BLUE: '#1C01CE',
        Color.CYAN: '#3F6E75',
        Color.GREEN: '#177500',
        Color.INTENSE_BLACK: '#3F6E75',
        Color.INTENSE_BLUE: '#2300CE',
        Color.INTENSE_CYAN: '#3F6E75',
        Color.INTENSE_GREEN: '#3F6E75',
        Color.INTENSE_PURPLE: '#A90D91',
        Color.INTENSE_RED: '#C41A16',
        Color.INTENSE_WHITE: '#3F6E75',
        Color.INTENSE_YELLOW: '#836C28',
        Color.NO_COLOR: 'noinherit',
        Color.PURPLE: '#A90D91',
        Color.RED: '#C41A16',
        Color.WHITE: '#3F6E75',
        Color.YELLOW: '#836C28',
        }
    _expand_style(style)
    return style


STYLES = LazyDict({
    'algol': _algol_style,
    'algol_nu': _algol_nu_style,
    'autumn': _autumn_style,
    'borland': _borland_style,
    'bw': _bw_style,
    'colorful': _colorful_style,
    'default': _default_style,
    'emacs': _emacs_style,
    'friendly': _friendly_style,
    'fruity': _fruity_style,
    'igor': _igor_style,
    'lovelace': _lovelace_style,
    'manni': _manni_style,
    'monokai': _monokai_style,
    'murphy': _murphy_style,
    'native': _native_style,
    'paraiso-dark': _paraiso_dark_style,
    'paraiso-light': _paraiso_light_style,
    'pastie': _pastie_style,
    'perldoc': _perldoc_style,
    'rrt': _rrt_style,
    'tango': _tango_style,
    'trac': _trac_style,
    'vim': _vim_style,
    'vs': _vs_style,
    'xcode': _xcode_style,
    }, globals(), 'STYLES')

del (_algol_style, _algol_nu_style, _autumn_style, _borland_style, _bw_style,
     _colorful_style, _default_style, _emacs_style, _friendly_style,
     _fruity_style, _igor_style, _lovelace_style, _manni_style, _monokai_style,
     _murphy_style, _native_style, _paraiso_dark_style, _paraiso_light_style,
     _pastie_style, _perldoc_style, _rrt_style, _tango_style, _trac_style,
     _vim_style, _vs_style, _xcode_style)


# dynamic syles
def make_pygments_style(pallette):
    """Makes a pygments style based on a color pallete."""
    global Color
    style = {getattr(Color, 'NO_COLOR'): 'noinherit'}
    for name, t in BASE_XONSH_COLORS.items():
        color = find_closest_color(t, pallette)
        style[getattr(Color, name)] = '#' + color
        style[getattr(Color, 'BOLD_'+name)] = 'bold #' + color
        style[getattr(Color, 'UNDERLINE_'+name)] = 'underline #' + color
        style[getattr(Color, 'BOLD_UNDERLINE_'+name)] = 'bold underline #' + color
        style[getattr(Color, 'BACKGROUND_'+name)] = 'bg:#' + color
    return style


def pygments_style_by_name(name):
    """Gets or makes a pygments color style by its name."""
    if name in STYLES:
        return STYLES[name]
    pstyle = get_style_by_name(name)
    pallette = make_pallete(pstyle.styles.values())
    astyle = make_pygments_style(pallette)
    STYLES[name] = astyle
    return astyle


#
# Formatter
#

@lazyobject
def XonshTerminal256Formatter():
    class XonshTerminal256FormatterProxy(terminal256.Terminal256Formatter):
        """Proxy class for xonsh terminal256 formatting that understands.
        xonsh color tokens.
        """

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # just keep the opening token for colors.
            color_names = set(map(str, Color.subtypes))
            for name, (opener, closer) in self.style_string.items():
                if name in color_names:
                    self.style_string[name] = (opener, '')
            # special case NO_COLOR, because it is special.
            self.style_string['Token.Color.NO_COLOR'] = ('\x1b[39m', '')

    return XonshTerminal256FormatterProxy



# -*- coding: utf-8 -*-
"""Prompt formatter for current jobs"""

import xonsh.jobs as xj


def _current_job():
    j = xj.get_next_task()
    if j is not None:
        if not j['bg']:
            cmd = j['cmds'][-1]
            s = cmd[0]
            if s == 'sudo' and len(cmd) > 1:
                s = cmd[1]
            return s



"""An implementation of yes for xonsh."""


def yes(args, stdin, stdout, stderr):
    """A yes command."""
    if '--help' in args:
        print(YES_HELP, file=stdout)
        return 0

    to_print = ["y"] if len(args) == 0 else [str(i) for i in args]

    while True:
        print(*to_print, file=stdout)

    return 0


YES_HELP = """Usage: yes [STRING]...
  or:  yes OPTION
Repeatedly output a line with all specified STRING(s), or 'y'.

      --help     display this help and exit

This version of yes was written in Python for the xonsh project: http://xon.sh
Based on yes from GNU coreutils: http://www.gnu.org/software/coreutils/"""
